#!/usr/bin/env python

import re
import sys
import argparse

field_file = "nodes.ads"
kind_file = "iirs.ads"
node_file = "iirs.ads"
template_file = "iirs.adb.in"
meta_base_file = "nodes_meta"
prefix_name = "Iir_Kind_"
prefix_range_name = "Iir_Kinds_"
type_name = "Iir_Kind"
node_type = "Iir"
conversions = ['uc', 'pos', 'grp']


class FuncDesc:
    def __init__(self, name, fields, conv, acc,
                 pname, ptype, rname, rtype):
        self.name = name
        self.fields = fields  # List of physical fields used
        self.conv = conv
        self.acc = acc  # access: Chain, Chain_Next, Ref, Of_Ref, Maybe_Ref,
        #                 Forward_Ref, Maybe_Forward_Ref
        self.pname = pname  # Parameter mame
        self.ptype = ptype  # Parameter type
        self.rname = rname  # value name (for procedure)
        self.rtype = rtype  # value type


class NodeDesc:
    def __init__(self, name, format, fields, attrs):
        self.name = name
        self.format = format
        self.fields = fields  # {field: FuncDesc} dict, defined for all fields
        self.attrs = attrs    # A {attr: FuncDesc} dict
        self.order = []       # List of fields name, in order of appearance.


class line:
    def __init__(self, string, no):
        self.l = string
        self.n = no


class EndOfFile(Exception):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return "end of file " + self.filename


class linereader:
    def __init__(self, filename):
        self.filename = filename
        self.f = open(filename)
        self.lineno = 0
        self.l = ''

    def get(self):
        self.l = self.f.readline()
        if not self.l:
            raise EndOfFile(self.filename)
        self.lineno = self.lineno + 1
        return self.l


class ParseError(Exception):
    def __init__(self, lr, msg):
        self.lr = lr
        self.msg = msg

    def __str__(self):
        return 'Error: ' + self.msg
        return 'Parse error at ' + self.lr.filname + ':' + self.lr.lineno + \
               ': ' + self.msg


# Return fields description.
# This is a dictionary.  The keys represent the possible format of a node.
# The values are dictionnaries representing fields.  Keys are fields name, and
# values are fields type.
def read_fields(file):
    fields = {}
    formats = []
    lr = linereader(file)

    #  Search for 'type Format_Type is'
    while lr.get() != '   type Format_Type is\n':
        pass

    # Skip '('
    if lr.get() != '     (\n':
        raise 'no open parenthesis after Format_Type'

    # Read formats
    l = lr.get()
    pat_field_name = re.compile('      Format_(\w+),?\n')
    while l != '     );\n':
        m = pat_field_name.match(l)
        if m is None:
            print l
            raise 'bad literal within Format_Type'
        name = m.group(1)
        formats.append(name)
        fields[name] = {}
        l = lr.get()

    # Read fields
    l = lr.get()
    pat_fields = re.compile('   -- Fields of Format_(\w+):\n')
    pat_field_desc = re.compile('   --   (\w+) : (\w+).*\n')
    format_name = ''
    common_desc = {}

    # Read until common fields.
    while l != '   -- Common fields are:\n':
        l = lr.get()
    format_name = 'Common'
    nbr_formats = 0

    while True:
        # 1) Read field description
        l = lr.get()
        desc = common_desc.copy()
        while True:
            m = pat_field_desc.match(l)
            if m is None:
                break
            desc[m.group(1)] = m.group(2)
            l = lr.get()
            # print 'For: ' + format_name + ': ' + m.group(1)

        # 2) Disp
        if format_name == 'Common':
            common_desc = desc
        else:
            fields[format_name] = desc

        # 3) Read next format
        if l == '\n':
            if nbr_formats == len(fields):
                break
            else:
                l = lr.get()

        # One for a format
        m = pat_fields.match(l)
        if m is not None:
            format_name = m.group(1)
            if format_name not in fields:
                raise ParseError(
                    lr, 'Format ' + format_name + ' is unknown')
            nbr_formats = nbr_formats + 1
        else:
            raise ParseError(lr, 'unhandled format line')

    return (formats, fields)


# Read kinds and kinds ranges.
def read_kinds(filename):
    lr = linereader(filename)
    kinds = []
    #  Search for 'type Iir_Kind is'
    while lr.get() != '   type ' + type_name + ' is\n':
        pass
    # Skip '('
    if lr.get() != '     (\n':
        raise ParseError(
            lr, 'no open parenthesis after "type ' + type_name + '"')

    # Read literals
    pat_node = re.compile('      ' + prefix_name + '(\w+),?( +-- .*)?\n')
    pat_comment = re.compile('( +-- .*)?\n')
    while True:
        l = lr.get()
        if l == '     );\n':
            break
        m = pat_node.match(l)
        if m:
            kinds.append(m.group(1))
            continue
        m = pat_comment.match(l)
        if not m:
            raise ParseError(lr, 'Unknown line within kind declaration')

    # Check subtypes
    pat_subtype = re.compile('   subtype ' + r'(\w+) is '
                              + type_name + ' range\n')
    pat_first = re.compile('     ' + prefix_name + r'(\w+) ..\n')
    pat_last = re.compile('     ' + prefix_name + r'(\w+);\n')
    pat_middle = re.compile('   --' + prefix_name + r'(\w+)\n')
    kinds_ranges = {}
    while True:
        l = lr.get()
        # Start of methods is also end of subtypes.
        if l == '   -- General methods.\n':
            break
        # Found a subtype.
        m = pat_subtype.match(l)
        if m:
            # Check first bound
            name = m.group(1)
            if not name.startswith(prefix_range_name):
                raise ParseError(lr, 'incorrect prefix for subtype')
            name = name[len(prefix_range_name):]
            l = lr.get()
            mf = pat_first.match(l)
            if not mf:
                raise ParseError(lr, 'badly formated first bound of subtype')
            first = kinds.index(mf.group(1))
            idx = first
            has_middle = None
            # Read until last bound
            while True:
                l = lr.get()
                ml = pat_middle.match(l)
                if ml:
                    # Check element in the middle
                    n = ml.group(1)
                    if n not in kinds:
                        raise ParseError(
                            lr, "unknown kind " + n + " in subtype")
                    if kinds.index(n) != idx + 1:
                        raise ParseError(
                            lr, "missing " + kinds[idx + 1] + " in subtype")
                    has_middle = True
                    idx = idx + 1
                else:
                    # Check last bound
                    ml = pat_last.match(l)
                    if ml:
                        last = kinds.index(ml.group(1))
                        if last != idx + 1 and has_middle:
                            raise ParseError(
                                lr, "missing " + kinds[idx] + " in subtype")
                        break
                    raise ParseError(lr, "unhandled line in subtype")
            kinds_ranges[name] = kinds[first:last+1]
    return (kinds, kinds_ranges)

# Read functions
def read_methods(filename):
    lr = linereader(filename)
    funcs = []
    pat_field = re.compile(r'   --  Field: ([\w,]+)( \w+)?( \(\w+\))?\n')
    pat_conv = re.compile(r'^ \((\w+)\)$')
    pat_func = re.compile(
        r'   function Get_(\w+) \((\w+) : (\w+)\) return (\w+);\n')
    pat_proc = re.compile(
        r'   procedure Set_(\w+) \((\w+) : (\w+); (\w+) : (\w+)\);\n')
    pat_end = re.compile('end [A-Za-z.]+;\n')
    while True:
        l = lr.get()
        # Start of methods
        if l == '   -- General methods.\n':
            break
    while True:
        l = lr.get()
        if pat_end.match(l):
            break
        m = pat_field.match(l)
        if m:
            fields = m.group(1).split(',')
            # Extract access modifier
            acc = m.group(2)
            if acc:
                acc = acc.strip()
            # Extract conversion
            conv = m.group(3)
            if conv:
                mc = pat_conv.match(conv)
                if not mc:
                    raise ParseError(lr, 'conversion ill formed')
                conv = mc.group(1)
                if conv not in conversions:
                    raise ParseError(lr, 'unknown conversion ' + conv)
            else:
                conv = None
            if len(fields) > 1 and conv != 'grp':
                raise ParseError(lr, 'bad conversion for multiple fields')
            # Read function
            l = lr.get()
            mf = pat_func.match(l)
            if not mf:
                raise ParseError(
                    lr, 'function declaration expected after Field')
            # Read procedure
            l = lr.get()
            mp = pat_proc.match(l)
            if not mp:
                raise ParseError(
                    lr, 'procedure declaration expected after function')
            # Consistency check between function and procedure
            if mf.group(1) != mp.group(1):
                raise ParseError(lr, 'function and procedure name mismatch')
            if mf.group(2) != mp.group(2):
                raise ParseError(lr, 'parameter name mismatch with function')
            if mf.group(3) != mp.group(3):
                raise ParseError(lr, 'parameter type mismatch with function')
            if mf.group(4) != mp.group(5):
                raise ParseError(lr, 'result type mismatch with function')
            funcs.append(FuncDesc(mf.group(1), fields, conv, acc,
                                  mp.group(2), mp.group(3),
                                  mp.group(4), mp.group(5)))

    return funcs


# Read description for one node
# LR is the line reader.  NAMES is the list of (node name, format)
#  (one description may describe several nodes).
# A comment start at column 2 or 4 or later.
def read_nodes_fields(lr, names, fields, nodes, funcs_dict):
    pat_only = re.compile('   -- Only for ' + prefix_name + '(\w+):\n')
    pat_only_bad = re.compile ('   -- *Only for.*\n')
    pat_field = re.compile('   --   Get/Set_(\w+) \((Alias )?([\w,]+)\)\n')
    pat_comment = re.compile('   --(|  [^ ].*|    .*)\n')

    # Create nodes
    cur_nodes = []
    for (nm, fmt) in names:
        if fmt not in fields:
            raise ParseError(lr, 'unknown format "{}"'.format(fmt))
        n = NodeDesc(nm, fmt, {x: None for x in fields[fmt]}, {})
        nodes[nm] = n
        cur_nodes.append(n)

    # Skip comments
    l = lr.l
    while pat_comment.match(l):
        l = lr.get()

    # Look for fields
    while l != '\n':
        # Skip comments
        while pat_comment.match(l):
            l = lr.get()

        # Handle 'Only ...'
        m = pat_only.match(l)
        if m:
            only_nodes = []
            while True:
                name = m.group(1)
                n = nodes.get(name, None)
                if n is None:
                    raise ParseError(lr, 'node is unknown')
                if n not in cur_nodes:
                    raise ParseError(lr, 'node not currently described')
                only_nodes.append(n)
                l = lr.get()
                m = pat_only.match(l)
                if not m:
                    break
        else:
            # By default a field applies to all nodes.
            only_nodes = cur_nodes

        # Skip comments
        while pat_comment.match(l):
            l = lr.get()

        # Handle field: '--  Get/Set_FUNC (Alias? FIELD)'
        m = pat_field.match(l)
        if not m:
            if pat_only_bad.match(l):
                raise ParseError(lr, "misleading 'Only for' comment")
            else:
                raise ParseError(lr, 'bad line in node description')

        func = m.group(1)
        alias = m.group(2)
        fields = m.group(3).split(',')

        # Check the function exists and if the field is correct.
        if func not in funcs_dict:
            raise ParseError(lr, 'unknown function')
        func = funcs_dict[func]
        if func.fields != fields:
            raise ParseError(lr, 'fields mismatch')

        for c in only_nodes:
            for f in fields:
                if f not in c.fields:
                    raise ParseError(
                        lr, 'field ' + f + ' does not exist in node')
            if not alias:
                for f in fields:
                    if c.fields[f]:
                        raise ParseError(
                            lr, 'field ' + f + ' already used')
                    c.fields[f] = func
                    c.order.append(f)
            c.attrs[func.name] = func

        l = lr.get()


def read_nodes(filename, kinds, kinds_ranges, fields, funcs):
    """Read description for all nodes."""
    lr = linereader(filename)
    funcs_dict = {x.name: x for x in funcs}
    nodes = {}

    # Skip until start
    while lr.get() != '   -- Start of ' + type_name + '.\n':
        pass

    pat_decl = re.compile('   -- ' + prefix_name + '(\w+) \((\w+)\)\n')
    pat_decls = re.compile('   -- ' + prefix_range_name + '(\w+) \((\w+)\)\n')
    pat_comment_line = re.compile('   --+\n')
    pat_comment_box = re.compile('   --(  .*)?\n')
    while True:
        l = lr.get()
        if l == '   -- End of ' + type_name + '.\n':
            break
        if l == '\n':
            continue
        m = pat_decl.match(l)
        if m:
            # List of nodes being described by the current description.
            names = []

            # Declaration of the first node
            while True:
                name = m.group(1)
                if name not in kinds:
                    raise ParseError(lr, 'unknown node')
                fmt = m.group(2)
                names.append((name, fmt))
                if name in nodes:
                    raise ParseError(
                        lr, 'node {} already described'.format(name));
                # There might be several nodes described at once.
                l = lr.get()
                m = pat_decl.match(l)
                if not m:
                    break
            read_nodes_fields(lr, names, fields, nodes, funcs_dict)
            continue
        m = pat_decls.match(l)
        if m:
            # List of nodes being described by the current description.
            name = m.group(1)
            fmt = m.group(2)
            names = [(k, fmt) for k in kinds_ranges[name]]
            l = lr.get()
            read_nodes_fields(lr, names, fields, nodes, funcs_dict)
            continue
        if pat_comment_line.match(l) or pat_comment_box.match(l):
            continue
        raise ParseError(lr, 'bad line in node description')

    for k in kinds:
        if k not in nodes:
            raise ParseError(lr, 'no desription for "{}"'.format(k))
    return nodes


def gen_choices(choices):
    """Generate a choice 'when A | B ... Z =>' using elements of CHOICES."""
    is_first = True
    for c in choices:
        if is_first:
            print '        ',
            print 'when',
        else:
            print
            print '        ',
            print '  |',
        print prefix_name + c,
        is_first = False
    print '=>'


def gen_get_format(formats, nodes, kinds):
    """Generate the Get_Format function."""
    print '   function Get_Format (Kind : ' + type_name + ') ' + \
          'return Format_Type is'
    print '   begin'
    print '      case Kind is'
    for f in formats:
        choices = [k for k in kinds if nodes[k].format == f]
        gen_choices(choices)
        print '            return Format_' + f + ';'
    print '      end case;'
    print '   end Get_Format;'


def gen_subprg_header(decl):
    if len(decl) < 76:
        print decl + ' is'
    else:
        print decl
        print '   is'
    print '   begin'


def gen_assert(func):
    print '      pragma Assert (' + func.pname + ' /= Null_' + node_type + ');'
    cond = '(Has_' + func.name + ' (Get_Kind (' + func.pname + ')),'
    msg = '"no field ' + func.name + '");'
    if len(cond) < 60:
        print '      pragma Assert ' + cond
        print '                     ' + msg
    else:
        print '      pragma Assert'
        print '         ' + cond
        print '          ' + msg


def get_field_type(fields, f):
    for fld in fields.values():
        if f in fld:
            return fld[f]
    return None


def gen_get_set(func, nodes, fields):
    """Generate Get_XXX/Set_XXX subprograms for FUNC."""
    rtype = func.rtype
    # If the function needs several fields, it must be user defined
    if func.conv == 'grp':
        print '   type %s_Conv is record' % rtype
        for f in func.fields:
            print '      %s: %s;' % (f, get_field_type(fields, f))
        print '   end record;'
        print '   pragma Pack (%s_Conv);' % rtype
        print "   pragma Assert (%s_Conv'Size = %s'Size);" % (rtype, rtype)
        print
    else:
        f = func.fields[0]
        g = 'Get_' + f + ' (' + func.pname + ')'

    s = func.rname
    if func.conv:
        if func.conv == 'uc':
            field_type = get_field_type(fields, f)
            g = field_type + '_To_' + rtype + ' (' + g + ')'
            s = rtype + '_To_' + field_type + ' (' + s + ')'
        elif func.conv == 'pos':
            g = rtype + "'Val (" + g + ')'
            s = rtype + "'Pos (" + s + ')'

    subprg = '   function Get_' + func.name + ' (' + func.pname \
             + ' : ' + func.ptype + ') return ' + rtype
    if func.conv == 'grp':
        print subprg
        print '   is'
        print '      function To_%s is new Ada.Unchecked_Conversion' % \
            func.rtype
        print '         (%s_Conv, %s);' % (rtype, rtype)
        print '      Conv : %s_Conv;' % rtype
        print '   begin'
    else:
        gen_subprg_header(subprg)
    gen_assert(func)
    if func.conv == 'grp':
        for f in func.fields:
            print '      Conv.%s := Get_%s (%s);' % (f, f, func.pname)
        g = 'To_%s (Conv)' % rtype
    print '      return ' + g + ';'
    print '   end Get_' + func.name + ';'
    print

    subprg = '   procedure Set_' + func.name + ' (' \
             + func.pname + ' : ' + func.ptype + '; ' \
             + func.rname + ' : ' + func.rtype + ')'
    if func.conv == 'grp':
        print subprg
        print '   is'
        print '      function To_%s_Conv is new Ada.Unchecked_Conversion' % \
            func.rtype
        print '         (%s, %s_Conv);' % (rtype, rtype)
        print '      Conv : %s_Conv;' % rtype
        print '   begin'
    else:
        gen_subprg_header(subprg)
    gen_assert(func)
    if func.conv == 'grp':
        print '      Conv := To_%s_Conv (%s);' % (rtype, func.rname)
        for f in func.fields:
            print '      Set_%s (%s, Conv.%s);' % (f, func.pname, f)
    else:
        print '      Set_' + f + ' (' + func.pname + ', ' + s + ');'
    print '   end Set_' + func.name + ';'
    print


def funcs_of_node(n):
    return sorted([fv.name for fv in n.fields.values() if fv])


def gen_has_func_spec(name, suff):
    spec = '   function Has_' + name + ' (K : ' + type_name + ')'
    ret = ' return Boolean' + suff
    if len(spec) < 60:
        print spec + ret
    else:
        print spec
        print '     ' + ret


def do_disp_formats():
    for fmt in fields:
        print "Fields of Format_"+fmt
        fld = fields[fmt]
        for k in fld:
            print '  ' + k + ' (' + fld[k] + ')'


def do_disp_kinds():
    print "Kinds are:"
    for k in kinds:
        print '  ' + prefix_name + k


def do_disp_funcs():
    print "Functions are:"
    for f in funcs:
        s = '{0} ({1}: {2}'.format(f.name, f.field, f.rtype)
        if f.acc:
            s += ' acc:' + f.acc
        if f.conv:
            s += ' conv:' + f.conv
        s += ')'
        print s


def do_disp_types():
    print "Types are:"
    s = set([])
    for f in funcs:
        s |= set([f.rtype])
    for t in sorted(s):
        print '  ' + t


def do_disp_nodes():
    for k in kinds:
        v = nodes[k]
        print prefix_name + k + ' (' + v.format + ')'
        flds = [fk for fk, fv in v.fields.items() if fv]
        for fk in sorted(flds):
            print '  ' + fk + ': ' + v.fields[fk].name


def do_get_format():
    gen_get_format(formats, nodes)


def do_body():
    lr = linereader(template_file)
    while True:
        l = lr.get().rstrip()
        print l
        if l == '   --  Subprograms':
            gen_get_format(formats, nodes, kinds)
            print
            for f in funcs:
                gen_get_set(f, nodes, fields)
        if l[0:3] == 'end':
            break


def get_types():
    s = set([])
    for f in funcs:
        s |= set([f.rtype])
    return [t for t in sorted(s)]


def get_attributes():
    s = set([])
    for f in funcs:
        if f.acc:
            s |= set([f.acc])
    res = [t for t in sorted(s)]
    res.insert(0, 'None')
    return res


def gen_enum(prefix, vals):
    last = None
    for v in vals:
        if last:
            print last + ','
        last = prefix + v
    print last


def do_meta_specs():
    lr = linereader(meta_base_file + '.ads.in')
    types = get_types()
    while True:
        l = lr.get().rstrip()
        if l == '      --  TYPES':
            gen_enum('      Type_', types)
        elif l == '      --  FIELDS':
            gen_enum('      Field_', [f.name for f in funcs])
        elif l == '      --  ATTRS':
            gen_enum('      Attr_', get_attributes())
        elif l == '   --  FUNCS':
            for t in types:
                print '   function Get_' + t
                print '      (N : ' + node_type + '; F : Fields_Enum) return '\
                      + t + ';'
                print '   procedure Set_' + t
                print '      (N : ' + node_type + '; F : Fields_Enum; V: ' \
                      + t + ');'
                print
            for f in funcs:
                gen_has_func_spec(f.name, ';')
        elif l[0:3] == 'end':
            print l
            break
        else:
            print l


def do_meta_body():
    lr = linereader(meta_base_file + '.adb.in')
    while True:
        l = lr.get().rstrip()
        if l == '      --  FIELDS_TYPE':
            last = None
            for f in funcs:
                if last:
                    print last + ','
                last = '      Field_' + f.name + ' => Type_' + f.rtype
            print last
        elif l == '         --  FIELD_IMAGE':
            for f in funcs:
                print '         when Field_' + f.name + ' =>'
                print '            return "' + f.name.lower() + '";'
        elif l == '         --  IIR_IMAGE':
            for k in kinds:
                print '         when ' + prefix_name + k + ' =>'
                print '            return "' + k.lower() + '";'
        elif l == '         --  FIELD_ATTRIBUTE':
            for f in funcs:
                print '         when Field_' + f.name + ' =>'
                if f.acc:
                    attr = f.acc
                else:
                    attr = 'None'
                print '            return Attr_' + attr + ';'
        elif l == '      --  FIELDS_ARRAY':
            last = None
            nodes_types = [node_type,
                           node_type + '_List', node_type + '_Flist']
            for k in kinds:
                v = nodes[k]
                if last:
                    print last + ','
                last = None
                print '      --  ' + prefix_name + k
                # Get list of physical fields for V, in some order.
                if flag_keep_order:
                    flds = v.order
                else:
                    # First non Iir and no Iir_List.
                    flds = sorted([fk for fk, fv in v.fields.items()
                                   if fv and fv.rtype not in nodes_types])
                    # Then Iir and Iir_List in order of appearance
                    flds += (fv for fv in v.order
                             if v.fields[fv].rtype in nodes_types)
                # Print the corresponding node field, but remove duplicate due
                # to 'grp'.
                fldsn = []
                for fk in flds:
                    if last:
                        print last + ','
                    # Remove duplicate
                    fn = v.fields[fk].name
                    if fn not in fldsn:
                        last = '      Field_' + fn
                        fldsn.append(fn)
                    else:
                        last = None
            if last:
                print last
        elif l == '      --  FIELDS_ARRAY_POS':
            pos = -1
            last = None
            for k in kinds:
                v = nodes[k]
                # Create a set to remove duplicate for 'grp'.
                flds = set([fv.name for fk, fv in v.fields.items() if fv])
                pos += len(flds)
                if last:
                    print last + ','
                last = '      ' + prefix_name + k + ' => {}'.format(pos)
            print last
        elif l == '   --  FUNCS_BODY':
            # Build list of types
            s = set([])
            for f in funcs:
                s |= set([f.rtype])
            types = [t for t in sorted(s)]
            for t in types:
                print '   function Get_' + t
                print '      (N : ' + node_type + '; F : Fields_Enum) return '\
                      + t + ' is'
                print '   begin'
                print '      pragma Assert (Fields_Type (F) = Type_' + t + ');'
                print '      case F is'
                for f in funcs:
                    if f.rtype == t:
                        print '         when Field_' + f.name + ' =>'
                        print '            return Get_' + f.name + ' (N);'
                print '         when others =>'
                print '            raise Internal_Error;'
                print '      end case;'
                print '   end Get_' + t + ';'
                print
                print '   procedure Set_' + t
                print '      (N : ' + node_type + '; F : Fields_Enum; V: ' \
                      + t + ') is'
                print '   begin'
                print '      pragma Assert (Fields_Type (F) = Type_' + t + ');'
                print '      case F is'
                for f in funcs:
                    if f.rtype == t:
                        print '         when Field_' + f.name + ' =>'
                        print '            Set_' + f.name + ' (N, V);'
                print '         when others =>'
                print '            raise Internal_Error;'
                print '      end case;'
                print '   end Set_' + t + ';'
                print
            for f in funcs:
                gen_has_func_spec(f.name, ' is')
                choices = [k for k in kinds if f.name in nodes[k].attrs]
                if len(choices) == 0:
                    print '      pragma Unreferenced (K);'
                print '   begin'
                if len(choices) == 0:
                    print '      return False;'
                elif len(choices) == 1:
                    print '      return K = ' + prefix_name + choices[0] + ';'
                else:
                    print '      case K is'
                    gen_choices(choices)
                    print '            return True;'
                    print '         when others =>'
                    print '            return False;'
                    print '      end case;'
                print '   end Has_' + f.name + ';'
                print
        elif l[0:3] == 'end':
            print l
            break
        else:
            print l


actions = {'disp-nodes': do_disp_nodes,
           'disp-kinds': do_disp_kinds,
           'disp-formats': do_disp_formats,
           'disp-funcs': do_disp_funcs,
           'disp-types': do_disp_types,
           'get_format': do_get_format,
           'body': do_body,
           'meta_specs': do_meta_specs,
           'meta_body': do_meta_body}


def main():
    parser = argparse.ArgumentParser(description='Meta-grammar processor')
    parser.add_argument('action', choices=actions.keys(),
                        default='disp-nodes')
    parser.add_argument('--field-file', dest='field_file',
                        default='nodes.ads',
                        help='specify file which defines fields')
    parser.add_argument('--kind-file', dest='kind_file',
                        default='iirs.ads',
                        help='specify file which defines nodes kind')
    parser.add_argument('--node-file', dest='node_file',
                        default='iirs.ads',
                        help='specify file which defines nodes and methods')
    parser.add_argument('--template-file', dest='template_file',
                        default='iirs.adb.in',
                        help='specify template body file')
    parser.add_argument('--meta-basename', dest='meta_basename',
                        default='nodes_meta',
                        help='specify base name of meta files')
    parser.add_argument('--kind-type', dest='kind_type',
                        default='Iir_Kind',
                        help='name of kind type')
    parser.add_argument('--kind-prefix', dest='kind_prefix',
                        default='Iir_Kind_',
                        help='prefix for kind literals')
    parser.add_argument('--kind-range-prefix', dest='kind_range_prefix',
                        default='Iir_Kinds_',
                        help='prefix for kind subtype (range)')
    parser.add_argument('--node-type', dest='node_type',
                        default='Iir',
                        help='name of the node type')
    parser.add_argument('--keep-order', dest='flag_keep_order',
                        action='store_true',
                        help='keep field order of nodes')
    parser.set_defaults(flag_keep_order=False)
    args = parser.parse_args()

    # At some point, it would be simpler to create a class...
    global formats, fields, nodes, kinds, kinds_ranges, funcs

    global type_name, prefix_name, template_file, node_type, meta_base_file
    global prefix_range_name, flag_keep_order, kind_file

    type_name = args.kind_type
    prefix_name = args.kind_prefix
    prefix_range_name = args.kind_range_prefix
    template_file = args.template_file
    node_type = args.node_type
    meta_base_file = args.meta_basename
    flag_keep_order = args.flag_keep_order

    field_file = args.field_file
    kind_file = args.kind_file
    node_file = args.node_file

    try:
        (formats, fields) = read_fields(field_file)
        (kinds, kinds_ranges) = read_kinds(kind_file)
        funcs = read_methods(node_file)
        nodes = read_nodes(node_file, kinds, kinds_ranges, fields, funcs)

    except ParseError as e:
        print >> sys.stderr, e
        print >> sys.stderr, \
            "in {0}:{1}:{2}".format(e.lr.filename, e.lr.lineno, e.lr.l)
        sys.exit(1)

    f = actions.get(args.action, None)
    if not f:
        print >> sys.stderr, "Action {0} is unknown".format(args.action)
        sys.exit(1)
    f()


if __name__ == '__main__':
    main()



from libghdl import libghdl
from ctypes import c_int

Std_Logic_1164_Pkg = c_int.in_dll(
    libghdl, "vhdl__ieee__std_logic_1164__std_logic_1164_pkg")

# Get value
Std_Logic_Type = c_int.in_dll(
    libghdl, "vhdl__ieee__std_logic_1164__std_logic_type")

# Get value
Std_Logic_Vector_Type = c_int.in_dll(
    libghdl, "vhdl__ieee__std_logic_1164__std_logic_vector_type")

# Get value
Rising_Edge = c_int.in_dll(
    libghdl, "vhdl__ieee__std_logic_1164__rising_edge")

# Get value
Falling_Edge = c_int.in_dll(
    libghdl, "vhdl__ieee__std_logic_1164__falling_edge")



#!/usr/bin/env python

import argparse
import glob
import subprocess
import select
import os
import os.path
import time

NUMJOBS=4

class Job(object):
    def __init__(self, dirname, poll):
        self.dirname = dirname
        self.poll = poll
        self.out = ''

    def start(self):
        self.p = subprocess.Popen(
            ['./testsuite.sh'],
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT, close_fds=True,
            cwd=self.dirname)
        self.out_fd = self.p.stdout
        self.poll.extend([self.out_fd])

    def wait(self):
        self.poll.remove(self.out_fd)
        return self.p.wait()

def run(keep):
    # List of tests to run
    tests = glob.glob('*[0-9]')

    start_time = time.time()
    jobs = []
    poll = []
    failures = []
    nbr_tests = len(tests)
    nbr_run = 0
    nbr_err = 0
    while len(tests) != 0 or len(jobs) != 0:
        # Start as many jobs as possible
        if len(tests) > 0 and len(jobs) < NUMJOBS:
            test = tests.pop(0)
            j = Job(test, poll)
            jobs.append(j)
            print('Starting {}'.format(test))
            j.start()
        elif len(jobs) > 0:
            # Wait for output or end of job.
            assert len(poll) == NUMJOBS or len(tests) == 0
            res = select.select(poll, [], [])
            done = set()
            for fd in res[0]:
                d = os.read(fd.fileno(), 1024)
                if len(d) == 0:
                    # EOF => end of job.
                    for j in jobs:
                        if fd == j.out_fd:
                            done.add(j)
                            break
                else:
                    # Gather output
                    for j in jobs:
                        if fd == j.out_fd:
                            j.out += d
            for j in done:
                print('Finish: {}'.format(j.dirname))
                print(j.out)
                code = j.wait()
                if code != 0:
                    print('############### Error for {}'.format(j.dirname))
                    nbr_err += 1
                    failures.append(j.dirname)
                    if not keep:
                        tests = []
                jobs.remove(j)
                nbr_run += 1
    end_time = time.time()

    print('{}/{} tests run in {} sec, {} failures'.format(
        nbr_run, nbr_tests, end_time - start_time, nbr_err))
    if failures:
        print('Failure: {}'.format(failures))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="parallel test driver")
    parser.add_argument("-k", "--keep",
                        help="keep running", action='store_true')
    args = parser.parse_args()
    run(args.keep)




from libghdl import libghdl
from ctypes import c_int32

Flist_Type = c_int32

Ffirst = 0
Flast = libghdl.vhdl__flists__flast

Length = libghdl.vhdl__flists__length

Get_Nth_Element = libghdl.vhdl__flists__get_nth_element



#!/usr/bin/env python

"""Like pnodes but output for python"""

from __future__ import print_function
import pnodes
import re

libname = 'libghdl'


def print_enum(name, vals):
    print()
    print()
    print('class {0}:'.format(name))
    for n, k in enumerate(vals):
        if k == 'None':
            k = 'PNone'
        print('    {0} = {1}'.format(k, n))


def do_class_kinds():
    print_enum(pnodes.prefix_name.rstrip('_'), pnodes.kinds)

    print()
    print()
    print('class Iir_Kinds:')
    for k, v in pnodes.kinds_ranges.items():
        print('    {0} = ['.format(k))
        first = True
        for e in v:
            if first:
                first = False
            else:
                print(',')
            print('        Iir_Kind.{}'.format(e), end='')
        print(']')
        print()

def do_iirs_subprg():
    classname = 'vhdl__nodes'
    print()
    print('Get_Kind = {0}.{1}__get_kind'.format(libname, classname))
    print('Get_Location = {0}.{1}__get_location'.format(libname, classname))
    for k in pnodes.funcs:
        print()
        print('Get_{0} = {1}.{2}__get_{3}'.format(
            k.name, libname, classname, k.name.lower()))
        print()
        print('Set_{0} = {1}.{2}__set_{3}'.format(
            k.name, libname, classname, k.name.lower(), k.pname, k.rname))


def do_libghdl_elocations():
    classname = 'vhdl__elocations'
    print('from libghdl import libghdl')
    print()
    for k in pnodes.funcs:
        print()
        print('Get_{0} = {1}.{2}__get_{3}'.format(
            k.name, libname, classname, k.name.lower()))
        print()
        print('Set_{0} = {1}.{2}__set_{3}'.format(
            k.name, libname, classname, k.name.lower(), k.pname, k.rname))


def do_class_types():
    print_enum('types', pnodes.get_types())


def do_types_subprg():
    print()
    for k in pnodes.get_types():
        print()
        print('Get_{0} = {1}.vhdl__nodes_meta__get_{2}'.format(
            k, libname, k.lower()))


def do_has_subprg():
    print()
    for f in pnodes.funcs:
        print()
        print('Has_{0} =\\'.format(f.name))
        print('    {0}.vhdl__nodes_meta__has_{1}'.format(libname, f.name.lower()))


def do_class_field_attributes():
    print_enum('Attr', ['ANone' if a == 'None' else a
                        for a in pnodes.get_attributes()])


def do_class_fields():
    print_enum('fields', [f.name for f in pnodes.funcs])


def read_enum(filename, type_name, prefix, class_name, g=lambda m:m.group(1)):
    """Read an enumeration declaration from :param filename:"""
    pat_decl = re.compile(r'   type {0} is$'.format(type_name))
    pat_enum = re.compile(r'      {0}(\w+),?( *-- .*)?$'.format(prefix))
    pat_comment = re.compile(r' *-- .*$')
    lr = pnodes.linereader(filename)
    while not pat_decl.match(lr.get()):
        pass
    line = lr.get()
    if line != '     (\n':
        raise pnodes.ParseError(
            lr, "{}:{}: missing open parenthesis".format(
                filename, lr.lineno))
    toks = []
    while True:
        line = lr.get()
        if line == '     );\n':
            break
        m = pat_enum.match(line)
        if m:
            toks.append(g(m))
        elif pat_comment.match(line):
            pass
        elif line == '\n':
            pass
        else:
            print(line, file=sys.stderr)
            raise pnodes.ParseError(
                lr, "{}:{}: incorrect line in enum {}".format(
                    filename, lr.lineno, type_name))
    print_enum(class_name, toks)


def read_spec_enum(type_name, prefix, class_name):
    """Read an enumeration declaration from iirs.ads"""
    read_enum(pnodes.kind_file, type_name, prefix, class_name)


def do_libghdl_nodes():
    print('from libghdl import libghdl')
    print("""
Null_Iir = 0

Null_Iir_List = 0
Iir_List_All = 1

Null_Iir_Flist = 0
Iir_Flist_Others = 1
Iir_Flist_All = 2
""")
    do_class_kinds()
    read_spec_enum('Iir_Mode', 'Iir_', 'Iir_Mode')
    read_spec_enum('Iir_Staticness', '', 'Iir_Staticness')
    read_spec_enum('Iir_Constraint', '', 'Iir_Constraint')
    read_spec_enum('Iir_Direction', 'Iir_', 'Iir_Direction')
    read_spec_enum('Iir_Delay_Mechanism', 'Iir_', 'Iir_Delay_Mechanism')
    read_spec_enum('Date_State_Type', 'Date_', 'Date_State')
    read_spec_enum('Iir_Predefined_Functions',
                   'Iir_Predefined_', 'Iir_Predefined')
    do_iirs_subprg()


def do_libghdl_meta():
    print('from libghdl import libghdl')
    print("""


# From nodes_meta
get_fields_first = libghdl.vhdl__nodes_meta__get_fields_first

get_fields_last = libghdl.vhdl__nodes_meta__get_fields_last

get_field_by_index = libghdl.vhdl__nodes_meta__get_field_by_index

get_field_type = libghdl.vhdl__nodes_meta__get_field_type

get_field_attribute = libghdl.vhdl__nodes_meta__get_field_attribute""")
    do_class_types()
    do_class_field_attributes()
    do_class_fields()
    do_types_subprg()
    do_has_subprg()


def do_libghdl_names():
    pat_name_first = re.compile(
        r'   Name_(\w+)\s+: constant Name_Id := (\d+);')
    pat_name_def = re.compile(
        r'   Name_(\w+)\s+:\s+constant Name_Id :=\s+Name_(\w+)( \+ (\d+))?;')
    dict = {}
    lr = pnodes.linereader('../std_names.ads')
    while True:
        line = lr.get()
        m = pat_name_first.match(line)
        if m:
            name_def = m.group(1)
            val = int(m.group(2))
            dict[name_def] = val
            res = [(name_def, val)]
            break
    val_max = 1
    while True:
        line = lr.get()
        if line == 'end Std_Names;\n':
            break
        if line.endswith(':=\n'):
            line = line.rstrip() + lr.get()
        m = pat_name_def.match(line)
        if m:
            name_def = m.group(1)
            name_ref = m.group(2)
            val = m.group(3)
            if not val:
                val = 0
            val_ref = dict.get(name_ref, None)
            if not val_ref:
                raise pnodes.ParseError(
                    lr, "name {0} not found".format(name_ref))
            val = val_ref + int(val)
            val_max = max(val_max, val)
            dict[name_def] = val
            res.append((name_def, val))
    print('class Name:')
    for n, v in res:
        # Avoid clash with Python names
        if n in ['False', 'True', 'None']:
            n = 'N' + n
        print('    {0} = {1}'.format(n, v))


def do_libghdl_tokens():
    read_enum("vhdl-tokens.ads", "Token_Type", "Tok_", "Tok")


def do_libghdl_errorout():
    print('from libghdl import libghdl')
    print("""
Enable_Warning = libghdl.errorout__enable_warning
""")
    read_enum("../errorout.ads", "Msgid_Type", "(Msgid|Warnid)_", "Msgid",
              g=lambda m: m.group(1) + '_' + m.group(2))


pnodes.actions.update({'class-kinds': do_class_kinds,
                       'libghdl-nodes': do_libghdl_nodes,
                       'libghdl-meta': do_libghdl_meta,
                       'libghdl-names': do_libghdl_names,
                       'libghdl-tokens': do_libghdl_tokens,
                       'libghdl-elocs': do_libghdl_elocations,
                       'libghdl-errorout': do_libghdl_errorout})


pnodes.main()



from libghdl import libghdl

Enable_Warning = libghdl.errorout__enable_warning



class Msgid:
    Msgid_Note = 0
    Warnid_Library = 1
    Warnid_Missing_Xref = 2
    Warnid_Default_Binding = 3
    Warnid_Binding = 4
    Warnid_Port = 5
    Warnid_Reserved_Word = 6
    Warnid_Pragma = 7
    Warnid_Nested_Comment = 8
    Warnid_Directive = 9
    Warnid_Parenthesis = 10
    Warnid_Vital_Generic = 11
    Warnid_Delayed_Checks = 12
    Warnid_Body = 13
    Warnid_Specs = 14
    Warnid_Universal = 15
    Warnid_Port_Bounds = 16
    Warnid_Runtime_Error = 17
    Warnid_Delta_Cycle = 18
    Warnid_Shared = 19
    Warnid_Hide = 20
    Warnid_Unused = 21
    Warnid_Others = 22
    Warnid_Pure = 23
    Warnid_Analyze_Assert = 24
    Warnid_Static = 25
    Msgid_Warning = 26
    Msgid_Error = 27
    Msgid_Fatal = 28



from ctypes import c_int32, c_char_p, c_bool
from libghdl import libghdl

Replace_Text = libghdl.files_map__editor__replace_text_ptr
Replace_Text.argstype = [c_int32, c_int32, c_int32, c_int32, c_char_p, c_int32]
Replace_Text.restype = c_bool

Fill_Text = libghdl.files_map__editor__fill_text_ptr

Check_Buffer_Content = libghdl.files_map__editor__check_buffer_content

Copy_Source_File = libghdl.files_map__editor__copy_source_file



from libghdl import libghdl
from ctypes import c_bool

Flag_Concurrent_Stmts = c_bool.in_dll(
    libghdl, "vhdl__canon__canon_flag_concurrent_stmts")

Flag_Configurations = c_bool.in_dll(
    libghdl, "vhdl__canon__canon_flag_configurations")

Flag_Associations = c_bool.in_dll(
    libghdl, "vhdl__canon__canon_flag_associations")

Extract_Sequential_Statement_Chain_Sensitivity = \
    libghdl.vhdl__canon__canon_extract_sequential_statement_chain_sensitivity



from libghdl import libghdl
from ctypes import c_int32, c_char_p

Indent_String = libghdl.vhdl__formatters__indent_string

Allocate_Handle = libghdl.vhdl__formatters__allocate_handle

Get_Length = libghdl.vhdl__formatters__get_length
Get_Length.restype = c_int32

Get_C_String = libghdl.vhdl__formatters__get_c_string
Get_C_String.restype = c_char_p

Free_Handle = libghdl.vhdl__formatters__free_handle



#!/usr/bin/env python
import libghdl
import libghdl.thin.name_table as name_table
import libghdl.thin.files_map as files_map
import libghdl.thin.vhdl.nodes as nodes
import libghdl.thin.vhdl.sem_lib as sem_lib

def init():
    """Initialization: set options and then load libaries"""
    libghdl.set_option(b'--std=08')
    libghdl.analyze_init()

def get_identifier_ptr(n):
    """Return the python string from node :param n: identifier"""
    return name_table.Get_Name_Ptr(nodes.Get_Identifier(n)).decode('utf-8')

def list_units(filename):
    # Load the file
    file_id = name_table.Get_Identifier(filename.encode('utf_8'))
    sfe = files_map.Read_Source_File(name_table.Null_Identifier, file_id)
    if sfe == files_map.No_Source_File_Entry:
        print("cannot open file '{}'".format(filename))
        return

    # Parse and analyze
    file = sem_lib.Load_File(sfe)

    # Display all design units
    unit = nodes.Get_First_Design_Unit(file)
    while unit != nodes.Null_Iir:
        lib_unit = nodes.Get_Library_Unit(unit)
        if nodes.Get_Kind(lib_unit) == nodes.Iir_Kind.Entity_Declaration:
            print('entity {}'.format(get_identifier_ptr(lib_unit)))
        elif nodes.Get_Kind(lib_unit) == nodes.Iir_Kind.Architecture_Body:
            print('architecture {}'.format(get_identifier_ptr(lib_unit)))
        else:
            print('unknown unit!')
        unit = nodes.Get_Chain(unit)


def main():
    init()
    list_units('demo.vhdl')

if __name__ == '__main__':
    main()



from os.path import dirname, join
import json

# Try to load JSON data from a file. If not found, use the argument as a tag name and retrieve the data from GitHub.
def getJSON(tag='all'):
   f = tag
   tag = '/'+tag
   if f == 'all':
      f = 'releases'
      tag = ''

   try:
      d = json.loads(open(join(dirname(__file__), f+'.json'), 'r').read())
   except:
      from urllib.request import urlopen
      d = json.loads(urlopen('https://api.github.com/repos/ghdl/ghdl/releases'+tag).read())
      json.dump(d, open(f+'.json', 'w'), indent=4)
   return d

#
# Functions to print table with format `[ [], [], [], ... ]` to reStructuredText
#

# Print a row of data elements.
def printTabRow(l, r):
   printTabItem(l, r, '| ')

# Print a rule. Two characters in 'b' define the type of rule. Expected values are '+-', '+=' or '| '.
def printTabRule(l, b):
   printTabItem(l, [b[1] for x in range(len(l))], b)

# Print a full row, be it a rule or data.
# Extend the width of each field to the size specified in 'l'.
def printTabItem(l, a, b):
   for y, z in enumerate(a):
      print((b + z).ljust(l[y]+3, b[1]), end='')
   print(b[0])

# Get number of cols from number of elements in row 0.
# Compute minimum number of characters required for each col.
def getTabColLens(t):
   cl = [0 for _ in t[0]]
   for row in t:
      for y, z in enumerate(row):
         cl[y] = max(cl[y], len(z))
   return cl

# Print a table using the functions above.
# The first row contains the headers.
def printTab(t):
   clens = getTabColLens(t)

   printTabRule(clens, '+-')
   printTabRow(clens, t[0])
   printTabRule(clens, '+=')
   for x in t[1:]:
      printTabRow(clens, x)
      printTabRule(clens, '+-')
   print()

#
# Print two versions of each shield. Onee for 'html' (`image::`) and one for 'latex' (`replace::`)
#

# Strip all non-alphanumeric characters when creating the labels
def stripLabel(label):
   import re
   pattern = re.compile('[\W_]+')
   return pattern.sub('', label)

def printShieldSrc(label, alt, img, target, latex=False):
   if latex:
      i = stripLabel(label)
      if label[-6:] == '/total':
         label = label[:-6]
      print('.. |l' + i + '| replace:: `' + label + '`_')
      print('.. _' + label + ': ' + target + '\n')
   else:
      print('.. |' + label + '| image:: '+ img + '\n',
            '   :target: ' + target + '\n',
            '   :height: 22\n',
            '   :alt: ' + alt + '\n')

#
# Display better OS and Backend names than those represented in the tarball name
#

def prettyOS(i):
   if i == 'fedora28':
      return 'Fedora 28'
   elif i == 'macosx':
      return 'Max OS X'
   elif i == 'mingw32':
      return 'Windows x86 (MinGW32)'
   elif  i == 'mingw64':
      return 'Windows x86 (MinGW64)'
   elif  i == 'stretch':
      return 'Debian 9 (Stretch)'
   elif  i == 'gpl':
      return 'Debian 9 (Stretch) GPL'
   elif  i == 'ubuntu14':
      return '14.04 LTS (Trusty Tahr)'
   return i

def prettyBackend(i):
   if i == 'llvm':
      return 'LLVM'
   if i == 'llvm-3.8':
      return 'LLVM (3.8)'
   return i

#
# Get, extract and process JSON data to create the shields and table with the assets of a release
#

def createTagShields(data='latest'):
   if isinstance(data, str):
      data = getJSON(data)

   assets=[['OS', 'Backend', 'Size', 'Downloads']]
   tag = data['tag_name']
   for x in data['assets']:
      name = x['name']
      s = []

      p = 'ghdl-gpl-'+tag[1:]
      if name[0:len(p)] == p:
         s = ['gpl', 'mcode']

      p = 'ghdl-'+tag[1:]
      if name[0:len(p)] == p:
         s = name[len(p)+1:-4].split('-',1)

      if len(s) > 1:
         assets.append([
            prettyOS(s[0]),
            prettyBackend(s[1]),
            (str(round(x['size']/1024**2, 2))+' MB').rjust(8),
            '|' + tag + '/' + name + '|']
         )

   assets.append(['Sum:', '', '', '|'+tag+'/total|'])

   for x in assets[1:-1]:
      i = x[3][1:-1]
      for latex in [False, True]:
         printShieldSrc(i, i,
            'https://img.shields.io/github/downloads/ghdl/ghdl/' + i + '.svg?longCache=true&style=flat-square&logo=github&label=%7F',
            'https://github.com/ghdl/ghdl/releases/download/' + i, latex=latex)

   return assets

# TODO: Is github.com/ghdl/ghdl/releases/download/<tag>/<file> subject to rate limit? Is there an alternative (not documented) domain?


#
# Get, extract and process JSON data to create the shields and list/table with all the releases except the latest
#

def createReleasesShields(tag='latest'):
   d = getJSON()
   from dateutil.parser import parse as parseTime
   releases = [['Date', 'Downloads']]
   if tag == 'latest':
      t = d[0]
   for x in d:
      name = x['tag_name']
      if tag == name:
         t = x
      date = parseTime( x['published_at'] ).strftime("%Y-%m-%d")
      releases.append([date, '|'+name+'/total|'])
      i = name
      for l in [False, True]:
         printShieldSrc(i+'/total', i+' Total',
            'https://img.shields.io/github/downloads/ghdl/ghdl/' + i + '/total.svg?longCache=true&style=flat-square&logo=github&label=%7F',
            'https://github.com/ghdl/ghdl/releases/' + i, l)

   out = {'releases': releases, 'assets': createTagShields(t)}
   import json
   json.dump(out, open('data.json', 'w'), indent=4)
   return out


#
# Print the table with all the assets of a release/tag
#

def printReleaseTab(assets, latex=False):
   if isinstance(assets, str):
      assets = getJSON(assets)["assets"]

   if latex:
      for y, z in enumerate(assets[1:]):
         assets[y+1] = z[0:3] + ['|l' + stripLabel(z[3][1:-1]) + '|']

   printTab(assets)

#
# Print list of releases, except the latest (second row)
#

def printReleasesList(releases, latex=False):
   if isinstance(releases, str):
      releases = getJSON(releases)["releases"]

   rs = [releases[0]]
   rs.extend(releases[2:])

   if latex:
     rs[0] = ['Release/Tag'] + [rs[0][0]]
     for x, r in enumerate(rs[1:]):
        rs[x+1] = ['|l' + stripLabel(r[1][1:-1]) + '|'] + [r[0]]
   else:
     rs[0] = ['Release/Tag'] + rs[0]
     for x, r in enumerate(rs[1:]):
        rs[x+1] = [r[1][1:-7]] + r

   printTab(rs)

#
# Create shields/badges from JSON file
#

def createShields(file='shields'):
   shields = getJSON(file)
   for k, v in shields.items():
      t = v['target']
      if t[0:3] == 'gh:':
        t = 'https://github.com/' + t[3:]

      printShieldSrc(
         'SHIELD:'+k,
         v['alt'],
         'https://img.shields.io/' + v['src'] + '&style=flat-square&longCache=true',
         t,
         False
      )



from libghdl import libghdl
from ctypes import c_int32

# Use .value
Std_Location = c_int32.in_dll(libghdl, "vhdl__std_package__std_location")

# Use .value
Standard_Package = c_int32.in_dll(
    libghdl, "vhdl__std_package__standard_package")

# Use .value
Character_Type_Definition = c_int32.in_dll(
    libghdl, "vhdl__std_package__character_type_definition")



from libghdl import libghdl

Strip_Denoting_Name = libghdl.vhdl__utils__strip_denoting_name

Get_Entity = libghdl.vhdl__utils__get_entity

Is_Second_Subprogram_Specification = \
    libghdl.vhdl__utils__is_second_subprogram_specification

Get_Entity_From_Entity_Aspect = \
    libghdl.vhdl__utils__get_entity_from_entity_aspect

Get_Interface_Of_Formal = \
    libghdl.vhdl__utils__get_interface_of_formal



class Name:
    First_Character = 1
    Last_Character = 256
    First_Keyword = 257
    Mod = 257
    Rem = 258
    Abs = 259
    Not = 260
    Access = 261
    After = 262
    Alias = 263
    All = 264
    Architecture = 265
    Array = 266
    Assert = 267
    Attribute = 268
    Begin = 269
    Block = 270
    Body = 271
    Buffer = 272
    Bus = 273
    Case = 274
    Component = 275
    Configuration = 276
    Constant = 277
    Disconnect = 278
    Downto = 279
    Else = 280
    Elsif = 281
    End = 282
    Entity = 283
    Exit = 284
    File = 285
    For = 286
    Function = 287
    Generate = 288
    Generic = 289
    Guarded = 290
    If = 291
    In = 292
    Inout = 293
    Is = 294
    Label = 295
    Library = 296
    Linkage = 297
    Loop = 298
    Map = 299
    New = 300
    Next = 301
    Null = 302
    Of = 303
    On = 304
    Open = 305
    Others = 306
    Out = 307
    Package = 308
    Port = 309
    Procedure = 310
    Process = 311
    Range = 312
    Record = 313
    Register = 314
    Report = 315
    Return = 316
    Select = 317
    Severity = 318
    Signal = 319
    Subtype = 320
    Then = 321
    To = 322
    Transport = 323
    Type = 324
    Units = 325
    Until = 326
    Use = 327
    Variable = 328
    Wait = 329
    When = 330
    While = 331
    With = 332
    And = 333
    Or = 334
    Xor = 335
    Nand = 336
    Nor = 337
    Last_Vhdl87 = 337
    Xnor = 338
    Group = 339
    Impure = 340
    Inertial = 341
    Literal = 342
    Postponed = 343
    Pure = 344
    Reject = 345
    Shared = 346
    Unaffected = 347
    Sll = 348
    Sla = 349
    Sra = 350
    Srl = 351
    Rol = 352
    Ror = 353
    Last_Vhdl93 = 353
    Protected = 354
    Last_Vhdl00 = 354
    Assume = 355
    Context = 356
    Cover = 357
    Default = 358
    Parameter = 359
    Property = 360
    Restrict = 361
    Restrict_Guarantee = 362
    Sequence = 363
    Vmode = 364
    Vprop = 365
    Vunit = 366
    Last_Vhdl08 = 366
    First_Ams_Keyword = 367
    Across = 367
    Break = 368
    Limit = 369
    Nature = 370
    Noise = 371
    Procedural = 372
    Quantity = 373
    Reference = 374
    Spectrum = 375
    Subnature = 376
    Terminal = 377
    Through = 378
    Tolerance = 379
    Last_AMS_Vhdl = 379
    Last_Keyword = 379
    First_Verilog = 380
    Always = 380
    Assign = 381
    Buf = 382
    Bufif0 = 383
    Bufif1 = 384
    Casex = 385
    Casez = 386
    Cmos = 387
    Deassign = 388
    Defparam = 389
    Disable = 390
    Edge = 391
    Endcase = 392
    Endfunction = 393
    Endmodule = 394
    Endprimitive = 395
    Endspecify = 396
    Endtable = 397
    Endtask = 398
    Force = 399
    Forever = 400
    Fork = 401
    Highz0 = 402
    Highz1 = 403
    Ifnone = 404
    Initial = 405
    Input = 406
    Join = 407
    Large = 408
    Macromodule = 409
    Medium = 410
    Module = 411
    Negedge = 412
    Nmos = 413
    Notif0 = 414
    Notif1 = 415
    Output = 416
    Pmos = 417
    Posedge = 418
    Primitive = 419
    Pull0 = 420
    Pull1 = 421
    Pulldown = 422
    Pullup = 423
    Realtime = 424
    Release = 425
    Reg = 426
    Repeat = 427
    Rcmos = 428
    Rnmos = 429
    Rpmos = 430
    Rtran = 431
    Rtranif0 = 432
    Rtranif1 = 433
    Scalared = 434
    Small = 435
    Specify = 436
    Specparam = 437
    Strong0 = 438
    Strong1 = 439
    Supply0 = 440
    Supply1 = 441
    Tablex = 442
    Task = 443
    Tran = 444
    Tranif0 = 445
    Tranif1 = 446
    Tri = 447
    Tri0 = 448
    Tri1 = 449
    Triand = 450
    Trior = 451
    Trireg = 452
    Vectored = 453
    Wand = 454
    Weak0 = 455
    Weak1 = 456
    Wire = 457
    Wor = 458
    Last_Verilog = 458
    First_V2001 = 459
    Automatic = 459
    Endgenerate = 460
    Genvar = 461
    Localparam = 462
    Unsigned = 463
    Signed = 464
    Last_V2001 = 464
    Uwire = 465
    First_SV3_0 = 466
    Always_Comb = 466
    Always_Ff = 467
    Always_Latch = 468
    Bit = 469
    Byte = 470
    Changed = 471
    Char = 472
    Const = 473
    Continue = 474
    Do = 475
    Endinterface = 476
    Endtransition = 477
    Enum = 478
    Export = 479
    Extern = 480
    Forkjoin = 481
    Iff = 482
    Import = 483
    Int = 484
    Interface = 485
    Logic = 486
    Longint = 487
    Longreal = 488
    Modport = 489
    Packed = 490
    Priority = 491
    Shortint = 492
    Shortreal = 493
    Static = 494
    Struct = 495
    Timeprecision = 496
    Timeunit = 497
    Transition = 498
    Typedef = 499
    Union = 500
    Unique = 501
    Unique0 = 502
    Void = 503
    Last_SV3_0 = 503
    First_SV3_1 = 504
    Chandle = 504
    Class = 505
    Clocking = 506
    Constraint = 507
    Dist = 508
    Endclass = 509
    Endclocking = 510
    Endprogram = 511
    Endproperty = 512
    Endsequence = 513
    Extends = 514
    Final = 515
    First_Match = 516
    Inside = 517
    Intersect = 518
    Join_Any = 519
    Join_None = 520
    Local = 521
    Program = 522
    Rand = 523
    Randc = 524
    Ref = 525
    Solve = 526
    String = 527
    Super = 528
    This = 529
    Throughout = 530
    Var = 531
    Virtual = 532
    Wait_Order = 533
    Last_SV3_1 = 533
    First_SV3_1a = 534
    Covergroup = 534
    Coverpoint = 535
    Endgroup = 536
    Endpackage = 537
    Expect = 538
    Foreach = 539
    Ignore_Bins = 540
    Illegal_Bins = 541
    Matches = 542
    Randcase = 543
    Randsequence = 544
    Tagged = 545
    Wildcard = 546
    Last_SV3_1a = 546
    First_SV2009 = 547
    Implies = 547
    S_Until = 548
    S_Until_With = 549
    Until_With = 550
    Last_SV2009 = 550
    First_Operator = 551
    Op_Equality = 551
    Op_Inequality = 552
    Op_Less = 553
    Op_Less_Equal = 554
    Op_Greater = 555
    Op_Greater_Equal = 556
    Op_Plus = 557
    Op_Minus = 558
    Op_Mul = 559
    Op_Div = 560
    Op_Exp = 561
    Op_Concatenation = 562
    Op_Condition = 563
    Op_Match_Equality = 564
    Op_Match_Inequality = 565
    Op_Match_Less = 566
    Op_Match_Less_Equal = 567
    Op_Match_Greater = 568
    Op_Match_Greater_Equal = 569
    Last_Operator = 569
    First_Attribute = 570
    Base = 570
    Left = 571
    Right = 572
    High = 573
    Low = 574
    Pos = 575
    Val = 576
    Succ = 577
    Pred = 578
    Leftof = 579
    Rightof = 580
    Reverse_Range = 581
    Length = 582
    Delayed = 583
    Stable = 584
    Quiet = 585
    Transaction = 586
    Event = 587
    Active = 588
    Last_Event = 589
    Last_Active = 590
    Last_Value = 591
    Last_Attribute = 591
    First_Vhdl87_Attribute = 592
    Behavior = 592
    Structure = 593
    Last_Vhdl87_Attribute = 593
    First_Vhdl93_Attribute = 594
    Ascending = 594
    Image = 595
    Value = 596
    Driving = 597
    Driving_Value = 598
    Simple_Name = 599
    Instance_Name = 600
    Path_Name = 601
    Last_Vhdl93_Attribute = 601
    First_Vhdl08_Attribute = 602
    Element = 602
    Last_Vhdl08_Attribute = 602
    First_AMS_Attribute = 603
    Contribution = 603
    Dot = 604
    Integ = 605
    Above = 606
    Zoh = 607
    Ltf = 608
    Ztf = 609
    Ramp = 610
    Slew = 611
    Last_AMS_Attribute = 611
    First_Standard = 612
    Std = 612
    Standard = 613
    Boolean = 614
    NFalse = 615
    NTrue = 616
    Character = 617
    Severity_Level = 618
    Note = 619
    Warning = 620
    Error = 621
    Failure = 622
    Universal_Integer = 623
    Universal_Real = 624
    Convertible_Integer = 625
    Convertible_Real = 626
    Integer = 627
    Real = 628
    Time = 629
    Fs = 630
    Ps = 631
    Ns = 632
    Us = 633
    Ms = 634
    Sec = 635
    Min = 636
    Hr = 637
    Max = 638
    Delay_Length = 639
    Now = 640
    Natural = 641
    Positive = 642
    Bit_Vector = 643
    File_Open_Kind = 644
    Read_Mode = 645
    Write_Mode = 646
    Append_Mode = 647
    File_Open_Status = 648
    Open_Ok = 649
    Status_Error = 650
    Name_Error = 651
    Mode_Error = 652
    Foreign = 653
    Boolean_Vector = 654
    To_Bstring = 655
    To_Binary_String = 656
    To_Ostring = 657
    To_Octal_String = 658
    To_Hstring = 659
    To_Hex_String = 660
    Integer_Vector = 661
    Real_Vector = 662
    Time_Vector = 663
    Digits = 664
    Format = 665
    Unit = 666
    Domain_Type = 667
    Quiescent_Domain = 668
    Time_Domain = 669
    Frequency_Domain = 670
    Domain = 671
    Frequency = 672
    Last_Standard = 672
    First_Charname = 673
    Nul = 673
    Soh = 674
    Stx = 675
    Etx = 676
    Eot = 677
    Enq = 678
    Ack = 679
    Bel = 680
    Bs = 681
    Ht = 682
    Lf = 683
    Vt = 684
    Ff = 685
    Cr = 686
    So = 687
    Si = 688
    Dle = 689
    Dc1 = 690
    Dc2 = 691
    Dc3 = 692
    Dc4 = 693
    Nak = 694
    Syn = 695
    Etb = 696
    Can = 697
    Em = 698
    Sub = 699
    Esc = 700
    Fsp = 701
    Gsp = 702
    Rsp = 703
    Usp = 704
    Del = 705
    C128 = 706
    C129 = 707
    C130 = 708
    C131 = 709
    C132 = 710
    C133 = 711
    C134 = 712
    C135 = 713
    C136 = 714
    C137 = 715
    C138 = 716
    C139 = 717
    C140 = 718
    C141 = 719
    C142 = 720
    C143 = 721
    C144 = 722
    C145 = 723
    C146 = 724
    C147 = 725
    C148 = 726
    C149 = 727
    C150 = 728
    C151 = 729
    C152 = 730
    C153 = 731
    C154 = 732
    C155 = 733
    C156 = 734
    C157 = 735
    C158 = 736
    C159 = 737
    Last_Charname = 737
    First_Misc = 738
    Guard = 738
    Deallocate = 739
    File_Open = 740
    File_Close = 741
    Read = 742
    Write = 743
    Flush = 744
    Endfile = 745
    I = 746
    J = 747
    F = 748
    L = 749
    P = 750
    R = 751
    S = 752
    V = 753
    External_Name = 754
    Open_Kind = 755
    First = 756
    Last = 757
    Textio = 758
    Work = 759
    Text = 760
    To_String = 761
    Minimum = 762
    Maximum = 763
    Untruncated_Text_Read = 764
    Textio_Read_Real = 765
    Textio_Write_Real = 766
    Get_Resolution_Limit = 767
    Control_Simulation = 768
    Step = 769
    Index = 770
    Item = 771
    Uu_File_Uu = 772
    Uu_Line_Uu = 773
    Label_Applies_To = 774
    Return_Port_Name = 775
    Map_To_Operator = 776
    Type_Function = 777
    Built_In = 778
    NNone = 779
    Last_Misc = 779
    First_Ieee = 780
    Ieee = 780
    Std_Logic_1164 = 781
    Std_Ulogic = 782
    Std_Ulogic_Vector = 783
    Std_Logic = 784
    Std_Logic_Vector = 785
    Rising_Edge = 786
    Falling_Edge = 787
    VITAL_Timing = 788
    VITAL_Level0 = 789
    VITAL_Level1 = 790
    Numeric_Std = 791
    Numeric_Bit = 792
    Unresolved_Unsigned = 793
    Unresolved_Signed = 794
    Std_Logic_Arith = 795
    Std_Logic_Signed = 796
    Std_Logic_Unsigned = 797
    Std_Logic_Textio = 798
    To_Integer = 799
    To_Unsigned = 800
    To_Signed = 801
    Resize = 802
    Std_Match = 803
    Shift_Left = 804
    Shift_Right = 805
    Rotate_Left = 806
    Rotate_Right = 807
    To_Bit = 808
    To_Bitvector = 809
    To_Stdulogic = 810
    To_Stdlogicvector = 811
    To_Stdulogicvector = 812
    Is_X = 813
    Conv_Signed = 814
    Conv_Unsigned = 815
    Conv_Integer = 816
    Math_Real = 817
    Ceil = 818
    Round = 819
    Log2 = 820
    Sin = 821
    Cos = 822
    Last_Ieee = 822
    First_Synthesis = 823
    Allconst = 823
    Allseq = 824
    Anyconst = 825
    Anyseq = 826
    Last_Synthesis = 826
    First_Directive = 827
    Define = 827
    Endif = 828
    Ifdef = 829
    Ifndef = 830
    Include = 831
    Timescale = 832
    Undef = 833
    Protect = 834
    Begin_Protected = 835
    End_Protected = 836
    Key_Block = 837
    Data_Block = 838
    Line = 839
    Celldefine = 840
    Endcelldefine = 841
    Default_Nettype = 842
    Resetall = 843
    Last_Directive = 843
    First_Systask = 844
    Bits = 844
    D_Root = 845
    D_Unit = 846
    Last_Systask = 846
    First_SV_Method = 847
    Size = 847
    Insert = 848
    Delete = 849
    Pop_Front = 850
    Pop_Back = 851
    Push_Front = 852
    Push_Back = 853
    Name = 854
    Len = 855
    Substr = 856
    Exists = 857
    Atoi = 858
    Itoa = 859
    Find = 860
    Find_Index = 861
    Find_First = 862
    Find_First_Index = 863
    Find_Last = 864
    Find_Last_Index = 865
    Num = 866
    Randomize = 867
    Pre_Randomize = 868
    Post_Randomize = 869
    Srandom = 870
    Get_Randstate = 871
    Set_Randstate = 872
    Seed = 873
    State = 874
    Last_SV_Method = 874
    First_BSV = 875
    uAction = 875
    uActionValue = 876
    BVI = 877
    uC = 878
    uCF = 879
    uE = 880
    uSB = 881
    uSBR = 882
    Action = 883
    Endaction = 884
    Actionvalue = 885
    Endactionvalue = 886
    Ancestor = 887
    Clocked_By = 888
    Default_Clock = 889
    Default_Reset = 890
    Dependencies = 891
    Deriving = 892
    Determines = 893
    Enable = 894
    Ifc_Inout = 895
    Input_Clock = 896
    Input_Reset = 897
    Instance = 898
    Endinstance = 899
    Let = 900
    Match = 901
    Method = 902
    Endmethod = 903
    Numeric = 904
    Output_Clock = 905
    Output_Reset = 906
    Par = 907
    Endpar = 908
    Path = 909
    Provisos = 910
    Ready = 911
    Reset_By = 912
    Rule = 913
    Endrule = 914
    Rules = 915
    Endrules = 916
    Same_Family = 917
    Schedule = 918
    Seq = 919
    Endseq = 920
    Typeclass = 921
    Endtypeclass = 922
    Valueof = 923
    uValueof = 924
    Last_BSV = 924
    First_Comment = 925
    Psl = 925
    Pragma = 926
    Synthesis = 927
    Synopsys = 928
    Translate_Off = 929
    Translate_On = 930
    Last_Comment = 930
    First_PSL = 931
    A = 931
    Af = 932
    Ag = 933
    Ax = 934
    Abort = 935
    Assume_Guarantee = 936
    Before = 937
    Clock = 938
    E = 939
    Ef = 940
    Eg = 941
    Ex = 942
    Endpoint = 943
    Eventually = 944
    Fairness = 945
    Fell = 946
    Forall = 947
    G = 948
    Inf = 949
    Inherit = 950
    Never = 951
    Next_A = 952
    Next_E = 953
    Next_Event = 954
    Next_Event_A = 955
    Next_Event_E = 956
    Prev = 957
    Rose = 958
    Strong = 959
    W = 960
    Whilenot = 961
    Within = 962
    X = 963
    Last_PSL = 963
    First_Edif = 964
    Celltype = 974
    View = 975
    Viewtype = 976
    Direction = 977
    Contents = 978
    Net = 979
    Viewref = 980
    Cellref = 981
    Libraryref = 982
    Portinstance = 983
    Joined = 984
    Portref = 985
    Instanceref = 986
    Design = 987
    Designator = 988
    Owner = 989
    Member = 990
    Number = 991
    Rename = 992
    Userdata = 993
    Last_Edif = 993



from libghdl import libghdl
from ctypes import c_bool

Parse_Design_File = libghdl.vhdl__parse__parse_design_file

Flag_Parse_Parenthesis = c_bool.in_dll(
    libghdl, "vhdl__parse__flag_parse_parenthesis")



from ctypes import (c_char_p, c_int32, c_int, c_bool, sizeof, c_void_p, byref)
import libghdl.thin.name_table as name_table
import libghdl.thin.vhdl.nodes as nodes
import libghdl.thin.vhdl.nodes_meta as nodes_meta
import libghdl.thin.vhdl.lists as lists
import libghdl.thin.vhdl.flists as flists
from libghdl.thin.vhdl.nodes_meta import (Attr, types)

def name_image(nameid):
    return name_table.Get_Name_Ptr(nameid).decode('utf-8')

def _build_enum_image(cls):
    d = [e for e in dir(cls) if e[0] != '_']
    res = [None] * len(d)
    for e in d:
        res[getattr(cls, e)] = e
    return res


_fields_image = _build_enum_image(nodes_meta.fields)


def fields_image(idx):
    """String representation of field idx"""
    return _fields_image[idx]


_kind_image = _build_enum_image(nodes.Iir_Kind)


def kind_image(k):
    """String representation of Iir_Kind k"""
    return _kind_image[k]


_types_image = _build_enum_image(nodes_meta.types)


def types_image(t):
    """String representation of Nodes_Meta.Types t"""
    return _types_image[t]


_attr_image = _build_enum_image(nodes_meta.Attr)


def attr_image(a):
    """String representation of Nodes_Meta.Attr a"""
    return _attr_image[a]


def leftest_location(n):
    while True:
        if n == Null_Iir:
            return No_Location
        k = nodes.Get_Kind(n)
        if k == nodes.Iir_Kind.Array_Subtype_Definition:
            n = nodes.Get_Subtype_Type_Mark(n)
        else:
            return nodes.Get_Location(n)


def fields_iter(n):
    """Iterate on fields of node n"""
    if n == nodes.Null_Iir:
        return
    k = nodes.Get_Kind(n)
    first = nodes_meta.get_fields_first(k)
    last = nodes_meta.get_fields_last(k)
    for i in range(first, last + 1):
        yield nodes_meta.get_field_by_index(i)


def chain_iter(n):
    """Iterate of a chain headed by node n"""
    while n != nodes.Null_Iir:
        yield n
        n = nodes.Get_Chain(n)


def chain_to_list(n):
    """Convert a chain headed by node n to a python list"""
    return [e for e in chain_iter(n)]


def nodes_iter(n):
    """Iterate of all nodes of n, including n.
    Nodes are returned only once."""
    if n == nodes.Null_Iir:
        return
#    print 'nodes_iter for {0}'.format(n)
    yield n
    for f in fields_iter(n):
        typ = nodes_meta.get_field_type(f)
#        print ' {0}: field {1} (type: {2})'.format(
#            n, fields_image(f), types_image(typ))
        if typ == nodes_meta.types.Iir:
            attr = nodes_meta.get_field_attribute(f)
            if attr == Attr.ANone:
                for n1 in nodes_iter(nodes_meta.Get_Iir(n, f)):
                    yield n1
            elif attr == Attr.Chain:
                n2 = nodes_meta.Get_Iir(n, f)
                while n2 != nodes.Null_Iir:
                    for n1 in nodes_iter(n2):
                        yield n1
                    n2 = nodes.Get_Chain(n2)
            elif attr == Attr.Maybe_Ref:
                if not nodes.Get_Is_Ref(n, f):
                    for n1 in nodes_iter(nodes_meta.Get_Iir(n, f)):
                        yield n1
        elif typ == types.Iir_List:
            attr = nodes_meta.get_field_attribute(f)
            if attr == Attr.ANone:
                for n1 in list_iter(nodes_meta.Get_Iir_List(n, f)):
                    for n2 in nodes_iter(n1):
                        yield n2
        elif typ == types.Iir_Flist:
            attr = nodes_meta.get_field_attribute(f)
            if attr == Attr.ANone:
                for n1 in flist_iter(nodes_meta.Get_Iir_Flist(n, f)):
                    for n2 in nodes_iter(n1):
                        yield n2


def list_iter(lst):
    """Iterate of all element of Iir_List lst."""
    if lst <= nodes.Iir_List_All:
        return
    iter = lists.Iterate(lst)
    while lists.Is_Valid(byref(iter)):
        yield lists.Get_Element(byref(iter))
        lists.Next(byref(iter))


def flist_iter(lst):
    """Iterate of all element of Iir_List lst."""
    if lst <= nodes.Iir_Flist_All:
        return
    for i in range(flists.Flast(lst) + 1):
        yield flists.Get_Nth_Element(lst, i)


def declarations_iter(n):
    """Iterator on all declarations in n."""
    k = nodes.Get_Kind(n)
    if nodes_meta.Has_Generic_Chain(k):
        for n1 in chain_iter(nodes.Get_Generic_Chain(n)):
            yield n1
    if nodes_meta.Has_Port_Chain(k):
        for n1 in chain_iter(nodes.Get_Port_Chain(n)):
            yield n1
    if nodes_meta.Has_Interface_Declaration_Chain(k):
        for n1 in chain_iter(nodes.Get_Interface_Declaration_Chain(n)):
            yield n1
    if nodes_meta.Has_Declaration_Chain(k):
        for n1 in chain_iter(nodes.Get_Declaration_Chain(n)):
            k1 = nodes.Get_Kind(n1)
            if k1 in nodes.Iir_Kinds.Specification \
               or k1 == nodes.Iir_Kind.Use_Clause:
                # Not a declaration
                pass
            elif k1 == nodes.Iir_Kind.Signal_Attribute_Declaration:
                # Not a declaration
                pass
            elif k1 in [nodes.Iir_Kind.Type_Declaration,
                        nodes.Iir_Kind.Anonymous_Type_Declaration]:
                yield n1
                # Handle nested declarations: record elements, physical units,
                # enumeration literals...
                typ = nodes.Get_Type_Definition(n1)
                for n2 in declarations_iter(n1):
                    yield n2
            else:
                yield n1
                # There can be nested declarations (subprograms)
                for n2 in declarations_iter(n1):
                    yield n2
    if nodes_meta.Has_Concurrent_Statement_Chain(k):
        for n1 in chain_iter(nodes.Get_Concurrent_Statement_Chain(n)):
            for n2 in declarations_iter(n1):
                yield n2
    if nodes_meta.Has_Sequential_Statement_Chain(k):
        for n1 in chain_iter(nodes.Get_Sequential_Statement_Chain(n)):
            for n2 in declarations_iter(n1):
                yield n2
    if nodes_meta.Has_Parameter_Specification(k):
        yield nodes.Get_Parameter_Specification(n)
    if nodes_meta.Has_Generate_Statement_Body(k):
        for n1 in declarations_iter(nodes.Get_Generate_Statement_Body(n)):
            yield n1
    if nodes_meta.Has_Else_Clause(k):
        n1 = nodes.Get_Else_Clause(n)
        if n1 != Null_Iir:
            for n2 in declarations_iter(n1):
                yield n2
    if nodes_meta.Has_Generate_Else_Clause(k):
        n1 = nodes.Get_Generate_Else_Clause(n)
        if n1 != Null_Iir:
            for n2 in declarations_iter(n1):
                yield n2
    if nodes_meta.Has_Block_Header(k):
        n1 = nodes.Get_Block_Header(n)
        if n1 != Null_Iir:
            for n2 in declarations_iter(n1):
                yield n2
    # All these nodes are handled:
    if k in [nodes.Iir_Kind.Entity_Declaration,
             nodes.Iir_Kind.Architecture_Body,
             nodes.Iir_Kind.Package_Declaration,
             nodes.Iir_Kind.Package_Body,
             nodes.Iir_Kind.Process_Statement,
             nodes.Iir_Kind.Sensitized_Process_Statement,
             nodes.Iir_Kind.Concurrent_Assertion_Statement,
             nodes.Iir_Kind.Concurrent_Simple_Signal_Assignment,
             nodes.Iir_Kind.Concurrent_Selected_Signal_Assignment,
             nodes.Iir_Kind.Concurrent_Conditional_Signal_Assignment,
             nodes.Iir_Kind.Concurrent_Procedure_Call_Statement,
             nodes.Iir_Kind.Block_Statement,
             nodes.Iir_Kind.Block_Header,
             nodes.Iir_Kind.For_Generate_Statement,
             nodes.Iir_Kind.If_Generate_Statement,
             nodes.Iir_Kind.Generate_Statement_Body,
             nodes.Iir_Kind.Assertion_Statement,
             nodes.Iir_Kind.Wait_Statement,
             nodes.Iir_Kind.Simple_Signal_Assignment_Statement,
             nodes.Iir_Kind.Variable_Assignment_Statement,
             nodes.Iir_Kind.For_Loop_Statement,
             nodes.Iir_Kind.While_Loop_Statement,
             nodes.Iir_Kind.Case_Statement,
             nodes.Iir_Kind.Null_Statement,
             nodes.Iir_Kind.Exit_Statement,
             nodes.Iir_Kind.Next_Statement,
             nodes.Iir_Kind.Procedure_Call_Statement,
             nodes.Iir_Kind.Signal_Declaration,
             nodes.Iir_Kind.Constant_Declaration,
             nodes.Iir_Kind.Variable_Declaration,
             nodes.Iir_Kind.File_Declaration,
             nodes.Iir_Kind.Object_Alias_Declaration,
             nodes.Iir_Kind.Attribute_Declaration,
             nodes.Iir_Kind.Component_Declaration,
             nodes.Iir_Kind.Use_Clause,
             nodes.Iir_Kind.If_Statement,
             nodes.Iir_Kind.Elsif,
             nodes.Iir_Kind.Return_Statement,
             nodes.Iir_Kind.Type_Declaration,
             nodes.Iir_Kind.Anonymous_Type_Declaration,
             nodes.Iir_Kind.Subtype_Declaration,
             nodes.Iir_Kind.Function_Declaration,
             nodes.Iir_Kind.Function_Body,
             nodes.Iir_Kind.Procedure_Declaration,
             nodes.Iir_Kind.Procedure_Body,
             nodes.Iir_Kind.Component_Instantiation_Statement,
             ]:
        return
    assert False, "unknown node of kind {}".format(kind_image(k))


def concurrent_stmts_iter(n):
    """Iterator on concurrent statements in n."""
    k = nodes.Get_Kind(n)
    if k == nodes.Iir_Kind.Design_File:
        for n1 in chain_iter(nodes.Get_First_Design_Unit(n)):
            for n2 in concurrent_stmts_iter(n1):
                yield n2
    elif k == nodes.Iir_Kind.Design_Unit:
        for n1 in concurrent_stmts_iter(nodes.Get_Library_Unit(n)):
            yield n1
    elif k == nodes.Iir_Kind.Entity_Declaration \
         or k == nodes.Iir_Kind.Architecture_Body \
         or k == nodes.Iir_Kind.Block_Statement:
        for n1 in chain_iter(nodes.Get_Concurrent_Statement_Chain(n)):
            yield n1
            for n2 in concurrent_stmts_iter(n1):
                yield n2
    elif k == nodes.Iir_Kind.For_Generate_Statement:
        for n1 in concurrent_stmts_iter(nodes.Get_Generate_Statement_Body(n)):
            yield n1
    elif k == nodes.Iir_Kind.If_Generate_Statement:
        while n != Null_Iir:
            for n1 in concurrent_stmts_iter(
                    nodes.Get_Generate_Statement_Body(n)):
                yield n1
            n = nodes.Get_Generate_Else_Clause(n)
    elif k == nodes.Iir_Kind.Case_Generate_Statement:
        alt = nodes.Get_Case_Statement_Alternative_Chain(n)
        for n1 in chain_iter(alt):
            blk = nodes.Get_Associated_Block(n1)
            if blk != Null_Iir:
                for n2 in concurrent_stmts_iter(
                        nodes.Get_Generate_Statement_Body(n)):
                    yield n2


def constructs_iter(n):
    """Iterator on library unit, concurrent statements and declarations
       that appear directly within a declarative part."""
    if n == thin.Null_Iir:
        return
    k = nodes.Get_Kind(n)
    if k == nodes.Iir_Kind.Design_File:
        for n1 in chain_iter(nodes.Get_First_Design_Unit(n)):
            for n2 in constructs_iter(n1):
                yield n2
    elif k == nodes.Iir_Kind.Design_Unit:
        n1 = nodes.Get_Library_Unit(n)
        yield n1
        for n2 in constructs_iter(n1):
            yield n2
    elif k in [nodes.Iir_Kind.Entity_Declaration,
               nodes.Iir_Kind.Architecture_Body,
               nodes.Iir_Kind.Block_Statement,
               nodes.Iir_Kind.Generate_Statement_Body]:
        for n1 in chain_iter(nodes.Get_Declaration_Chain(n)):
            yield n1
            for n2 in constructs_iter(n1):
                yield n2
        for n1 in chain_iter(nodes.Get_Concurrent_Statement_Chain(n)):
            yield n1
            for n2 in constructs_iter(n1):
                yield n2
    elif k in [nodes.Iir_Kind.Configuration_Declaration,
               nodes.Iir_Kind.Package_Declaration,
               nodes.Iir_Kind.Package_Body,
               nodes.Iir_Kind.Function_Body,
               nodes.Iir_Kind.Procedure_Body,
               nodes.Iir_Kind.Protected_Type_Declaration,
               nodes.Iir_Kind.Protected_Type_Body,
               nodes.Iir_Kind.Process_Statement,
               nodes.Iir_Kind.Sensitized_Process_Statement]:
        for n1 in chain_iter(nodes.Get_Declaration_Chain(n)):
            yield n1
            for n2 in constructs_iter(n1):
                yield n2
    elif k == nodes.Iir_Kind.For_Generate_Statement:
        n1 = nodes.Get_Generate_Statement_Body(n)
        yield n1
        for n2 in constructs_iter(n1):
            yield n2
    elif k == nodes.Iir_Kind.If_Generate_Statement:
        while n != Null_Iir:
            n1 = nodes.Get_Generate_Statement_Body(n)
            yield n1
            for n2 in constructs_iter(n1):
                yield n2
            n = nodes.Get_Generate_Else_Clause(n)
    elif k == nodes.Iir_Kind.Case_Generate_Statement:
        alt = nodes.Get_Case_Statement_Alternative_Chain(n)
        for n1 in chain_iter(alt):
            blk = nodes.Get_Associated_Block(n1)
            if blk != Null_Iir:
                n2 = nodes.Get_Generate_Statement_Body(blk)
                yield n2
                for n3 in constructs_iter(n2):
                    yield n3

def sequential_iter(n):
    """Iterator on sequential statements.  The first node must be either
       a process or a subprogram body."""
    if n == thin.Null_Iir:
        return
    k = nodes.Get_Kind(n)
    if k in [nodes.Iir_Kind.Process_Statement,
             nodes.Iir_Kind.Sensitized_Process_Statement,
             nodes.Iir_Kind.Function_Body,
             nodes.Iir_Kind.Procedure_Body]:
        for n1 in chain_iter(nodes.Get_Sequential_Statement_Chain(n)):
            yield n1
            for n2 in sequential_iter(n1):
                yield n2
    elif k == nodes.Iir_Kind.If_Statement:
        while True:
            n = nodes.Get_Chain(n)
            if n == thin.Null_Iir:
                break
            yield n
            for n1 in sequential_iter(n):
                yield n1
    elif k == nodes.Iir_Kind.Case_Statement:
        for ch in chain_iter(nodes.Get_Case_Statement_Alternative_Chain(n)):
            stmt = nodes.Get_Associated_Chain(ch)
            if stmt != thin.Null_Iir:
                for n1 in chain_iter(stmt):
                    yield n1
                    for n2 in sequential_iter(n1):
                        yield n2
    elif k in [nodes.Iir_Kind.For_Loop_Statement,
               nodes.Iir_Kind.While_Loop_Statement]:
        for n1 in chain_iter(nodes.Get_Sequential_Statement_Chain(n)):
            yield n1
            for n2 in sequential_iter(n1):
                yield n2
    elif k in [nodes.Iir_Kind.Assertion_Statement,
               nodes.Iir_Kind.Wait_Statement,
               nodes.Iir_Kind.Null_Statement,
               nodes.Iir_Kind.Exit_Statement,
               nodes.Iir_Kind.Next_Statement,
               nodes.Iir_Kind.Return_Statement,
               nodes.Iir_Kind.Variable_Assignment_Statement,
               nodes.Iir_Kind.Simple_Signal_Assignment_Statement,
               nodes.Iir_Kind.Procedure_Call_Statement]:
        return
    else:
        assert False, "unknown node of kind {}".format(kind_image(k))



import os
import sys
from string import join
from vunit import VUnit



# Create VUnit instance by parsing command line arguments
vu = VUnit.from_argv()
vu.add_com()
lib = vu.add_library("lib")
lib.add_source_files("*.vhd")

vu.main()


from libghdl import libghdl
from ctypes import c_int32

Get_Libraries_Chain = libghdl.libraries__get_libraries_chain

Add_Design_Unit_Into_Library = \
    libghdl.libraries__add_design_unit_into_library

# Use .value
Library_Location = c_int32.in_dll(libghdl, "libraries__library_location")

# Use .value
Work_Library = c_int32.in_dll(libghdl, "libraries__work_library")

Purge_Design_File = libghdl.libraries__purge_design_file

Find_Entity_For_Component = libghdl.libraries__find_entity_for_component

Get_Library_No_Create = libghdl.libraries__get_library_no_create

Find_Primary_Unit = libghdl.libraries__find_primary_unit



#!/usr/bin/env python
#  Generate the body of ieee.std_logic_1164 from a template.
#  This file is part of GHDL.
#  Copyright (C) 2015 Tristan Gingold
#
#  GHDL is free software; you can redistribute it and/or modify it under
#  the terms of the GNU General Public License as published by the Free
#  Software Foundation; either version 2, or (at your option) any later
#  version.
#
#  GHDL is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or
#  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
#  for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with GCC; see the file COPYING2.  If not see
#  <http://www.gnu.org/licenses/>.

import re
import sys

# Supported versions
V87=0
V93=1
V08=2

binary_funcs = [ "and", "nand", "or", "nor", "xor" ]

# Python modelisation of std_ulogic type.
std_logic = "UX01ZWLH-"

# Normalization map.
ux01_map = { 'U': 'U',
             'X': 'X', 'Z': 'X', 'W': 'X', '-': 'X',
             '0': '0', 'L': '0',
             '1': '1', 'H': '1' }

def sl_and(a,b):
    """and definition"""
    na = ux01_map[a]
    nb = ux01_map[b]
    if na == '0' or nb == '0':
        return '0'
    if na == 'U' or nb == 'U':
        return 'U'
    if na == 'X' or nb == 'X':
        return 'X'
    return '1'

def sl_not(a):
    """not definition"""
    na = ux01_map[a]
    if na == 'U':
        return 'U'
    if na == 'X':
        return 'X'
    if na == '1':
        return '0'
    else:
        return '1'

def sl_or(a,b):
    "or definition"
    return sl_not(sl_and (sl_not(a), sl_not(b)))

def sl_xor(a,b):
    "xor definition"
    return sl_or(sl_and(a, sl_not(b)),
                 sl_and(sl_not(a), b))

# Stream to write.
out=sys.stdout

def w(s):
    "Write S to the output"
    out.write(s)

def gen_log_table2(name, func):
    "Generate a logic table for binary operator NAME using its model FUNC"
    w("""
  constant {0}_table : table_2d :=
  --  UX01ZWLH-
    (""".format(name))
    for a in std_logic:
        w('"')
        for b in std_logic:
            w(func (a, b))
        w('"')
        if a != '-':
            w(',')
        else:
            w(' ')
        w('   -- {}\n'.format(a))
        w('     ')
    w(');\n')

def gen_log_table1(name, func):
    "Generate a logic table for unary operator NAME using its model FUNC"
    w("""
  constant {0}_table : table_1d :=
  --  UX01ZWLH-
     """.format(name))
    w('"')
    for b in std_logic:
        w(func (b))
    w('";\n')

def disp_tables(version):
    "Generate logic tables"
    gen_log_table2("and", sl_and)
    gen_log_table2("nand", lambda a,b : sl_not(sl_and(a, b)))
    gen_log_table2("or", sl_or)
    gen_log_table2("nor", lambda a,b : sl_not(sl_or(a,b)))
    gen_log_table2("xor", sl_xor)
    if version >= V93:
        gen_log_table2("xnor", lambda a,b : sl_not(sl_xor(a,b)))
    gen_log_table1("not", sl_not)

vec_types = ['ulogic', 'logic']

def disp_scalar_binary(fun):
    "Generate scalar binary function body"
    w("""
  function "{0}" (l : std_ulogic; r : std_ulogic) return UX01 is
  begin
    return {0}_table (l, r);
  end "{0}";\n""".format(fun))

def disp_scalar_unary(fun):
    "Generate scalar unary function body"
    w("""
  function "{0}" (l : std_ulogic) return UX01 is
  begin
    return {0}_table (l);
  end "{0}";\n""".format(fun))

def disp_vec_binary(func, typ):
    "Generate vector binary function body"
    w("""
  function "{0}" (l, r : std_{1}_vector) return std_{1}_vector
  is
    subtype res_type is std_{1}_vector (1 to l'length);
    alias la : res_type is l;
    alias ra : std_{1}_vector (1 to r'length) is r;
    variable res : res_type;
  begin
    if la'length /= ra'length then
      assert false
        report "arguments of overloaded '{0}' operator are not of the same length"
        severity failure;
    else
      for I in res_type'range loop
        res (I) := {0}_table (la (I), ra (I));
      end loop;
    end if;
    return res;
  end "{0}";\n""".format(func, typ))

def disp_vec_unary(func, typ):
    "Generate vector unary function body"
    w("""
  function "{0}" (l : std_{1}_vector) return std_{1}_vector
  is
    subtype res_type is std_{1}_vector (1 to l'length);
    alias la : res_type is l;
    variable res : res_type;
  begin
    for I in res_type'range loop
      res (I) := {0}_table (la (I));
    end loop;
    return res;
  end "{0}";\n""".format(func, typ))

def disp_scal_vec_binary(func, typ):
    "Generate scalar-vector binary function body"
    w("""
  function "{0}" (l : std_{1}_vector; r : std_{1})
    return std_{1}_vector
  is
    subtype res_type is std_{1}_vector (1 to l'length);
    alias la : res_type is l;
    variable res : res_type;
  begin
    for I in res_type'range loop
      res (I) := {0}_table (la (I), r);
    end loop;
    return res;
  end "{0}";\n""".format(func, typ))

def disp_vec_scal_binary(func, typ):
    "Generate vector-scalar binary function body"
    w("""
  function "{0}" (l : std_{1}; r : std_{1}_vector)
    return std_{1}_vector
  is
    subtype res_type is std_{1}_vector (1 to r'length);
    alias ra : res_type is r;
    variable res : res_type;
  begin
    for I in res_type'range loop
      res (I) := {0}_table (l, ra (I));
    end loop;
    return res;
  end "{0}";\n""".format(func, typ))

def disp_vec_reduction(func, typ):
    "Generate reduction function body"
    init = '1' if func in ['and', 'nand'] else '0'
    w("""
  function "{0}" (l : std_{1}_vector) return std_{1}
  is
    variable res : std_{1} := '{2}';
  begin
    for I in l'range loop
      res := {0}_table (l(I), res);
    end loop;
    return res;
  end "{0}";\n""".format(func, typ, init))

def gen_shift(is_left, plus, minus):
    if is_left:
        s = "res (1 to l'length {1} r) := la (r {0} 1 to res'right);\n"
    else:
        s = "res (1 {0} r to res'right) := la (1 to l'length {1} r);\n"
    w("      " + s.format(plus, minus))

def disp_shift_funcs(func, typ):
    "Generate shift functions"
    w("""
  function "{0}" (l : std_{1}_vector; r : integer)
    return std_{1}_vector
  is
    subtype res_type is std_{1}_vector (1 to l'length);
    alias la : res_type is l;
    variable res : res_type := (others => '0');
  begin
    if r >= 0 then\n""".format(func, typ))
    gen_shift(func == "sll", '+', '-')
    w("    else\n")
    gen_shift(func != "sll", '-', '+')
    w("""    end if;
    return res;
  end "{0}";\n""".format(func, typ))

def gen_rot(is_left, plus, minus):
    if is_left:
        t = ["res (1 to res'right {1} rm) := la (rm {0} 1 to la'right);",
             "res (res'right {1} rm + 1 to res'right) := la (1 to rm);"]
    else:
        t = ["res (1 {0} rm to res'right) := la (1 to la'right {1} r);",
             "res (1 to rm) := la (la'right {1} rm + 1 to la'right);"]
    for s in t:
        w("      ")
        w(s.format(plus, minus))
        w('\n')

def disp_rot_funcs(func, typ):
    "Generate rotation functions"
    w("""
  function "{0}" (l : std_{1}_vector; r : integer)
    return std_{1}_vector
  is
    subtype res_type is std_{1}_vector (1 to l'length);
    alias la : res_type is l;
    variable res : res_type;
    constant rm : integer := r mod l'length;
  begin
    if r >= 0 then\n""".format(func, typ))
    gen_rot(func == "rol", '+', '-')
    w("    else\n")
    gen_rot(func != "rol", '-', '+')
    w("""    end if;
    return res;
  end "{0}";\n""".format(func, typ))

def disp_all_log_funcs(version):
    "Generate all function bodies for logic operators"
    for f in binary_funcs:
        disp_scalar_binary(f)
    disp_scalar_unary("not")
    for v in vec_types:
        typ = "std_" + v + "_vector"
        for f in binary_funcs:
            disp_vec_binary(f, v)
        disp_vec_unary("not", v)
        if version >= V08:
            for f in binary_funcs:
                disp_scal_vec_binary(f, v)
                disp_vec_scal_binary(f, v)
            for f in binary_funcs:
                disp_vec_reduction(f, v)
            disp_shift_funcs("sll", v)
            disp_shift_funcs("srl", v)
            disp_rot_funcs("rol", v)
            disp_rot_funcs("ror", v)

def disp_sv_to_bv_conv(typ):
    "Generate logic vector to bit vector function body"
    w("""
  function to_bitvector (s : std_{0}_vector; xmap : bit := '0')
    return bit_vector
  is
    subtype res_range is natural range s'length - 1 downto 0;
    alias as : std_{0}_vector (res_range) is s;
    variable res : bit_vector (res_range);
    variable b : bit;
  begin
    for I in res_range loop
      --  Inline for efficiency.
      case as (I) is
        when '0' | 'L' =>
          b := '0';
        when '1' | 'H' =>
          b := '1';
        when others =>
          b := xmap;
      end case;
      res (I) := b;
    end loop;
    return res;
  end to_bitvector;\n""".format(typ))

def disp_bv_to_sv_conv(typ):
    "Generate bit vector to logic vector function body"
    w("""
  function to_std{0}vector (b : bit_vector) return std_{0}_vector is
    subtype res_range is natural range b'length - 1 downto 0;
    alias ab : bit_vector (res_range) is b;
    variable res : std_{0}_vector (res_range);
  begin
    for I in res_range loop
      res (I) := bit_to_std (ab (I));
    end loop;
    return res;
  end to_std{0}vector;\n""".format(typ))

def disp_sv_to_sv_conv(s,d):
    "Generate logic vector to logic vector function body"
    w("""
  function to_std{1}vector (s : std_{0}_vector) return std_{1}_vector
  is
    subtype res_type is std_{1}_vector (s'length - 1 downto 0);
  begin
    return res_type (s);
  end to_std{1}vector;\n""".format(s,d))

def disp_all_conv_funcs(version):
    "Generate conversion function bodies"
    for v in vec_types:
        disp_sv_to_bv_conv(v)
    for v in vec_types:
        disp_bv_to_sv_conv(v)
    if version >= V08:
        disp_bv_to_sv_conv('logic')
    disp_sv_to_sv_conv('logic', 'ulogic')
    disp_sv_to_sv_conv('ulogic', 'logic')

def disp_conv_vec_vec(typ, v):
    "Generate function body for vector conversion"
    utyp = typ.upper();
    w("""
  function to_{1} (s : std_{2}_vector) return std_{2}_vector
  is
    subtype res_type is std_{2}_vector (1 to s'length);
    alias sa : res_type is s;
    variable res : res_type;
  begin
    for i in res_type'range loop
      res (i) := std_to_{0} (sa (i));
    end loop;
    return res;
  end to_{1};\n""".format(typ, utyp, v))

def disp_conv_std(typ):
    "Generate function body for scalar conversion"
    utyp = typ.upper();
    w("""
  function to_{1} (s : std_ulogic) return {1} is
  begin
    return std_to_{0} (s);
  end to_{1};\n""".format(typ, utyp))

def disp_conv_bv_vec(typ, v):
    "Generate function body for bit vector conversion"
    utyp = typ.upper();
    w("""
  function to_{1} (b : bit_vector) return std_{2}_vector
  is
    subtype res_range is natural range 1 to b'length;
    alias ba : bit_vector (res_range) is b;
    variable res : std_{2}_vector (res_range);
  begin
    for i in res_range loop
      res (i) := bit_to_x01 (ba (i));
    end loop;
    return res;
  end to_{1};\n""".format(typ, utyp, v))

def disp_conv_b_t(typ):
    "Generate function body for bit conversion"
    utyp = typ.upper();
    w("""
  function to_{1} (b : bit) return {1} is
  begin
    return bit_to_x01 (b);
  end to_{1};\n""".format(typ, utyp))

def disp_conv_01():
    "Generate to_01 bodies"
    w("""
  function to_01 (s : std_{0}_vector; xmap : std_ulogic := '0')
    return std_{0}_vector
  is
    subtype res_type is std_{0}_vector (s'length - 1 downto 0);
    alias sa : res_type is s;
    variable res : res_type;
  begin
    for i in res_type'range loop
      case sa(i) is
        when '0' | 'L' => res (i) := '0';
        when '1' | 'H' => res (i) := '1';
        when others    => return res_type'(others => xmap);
      end case;
    end loop;
    return res;
  end to_01;\n""".format("ulogic"))
    w("")
    w("""
  function to_01 (s : std_{0}; xmap : std_ulogic := '0')
    return std_{0} is
  begin
    case s is
      when '0' | 'L' => return '0';
      when '1' | 'H' => return '1';
      when others    => return xmap;
    end case;
  end to_01;\n""".format("ulogic"))
    w("")
    w("""
  function to_01 (s : bit_vector; xmap : std_ulogic := '0')
    return std_{0}_vector
  is
    alias sa : bit_vector(s'length - 1 downto 0) is s;
    variable res : std_{0}_vector (s'length - 1 downto 0);
  begin
    for i in sa'range loop
      res (i) := bit_to_std (sa (i));
    end loop;
    return res;
  end to_01;\n""".format("ulogic"))
    w("")
    w("""
  function to_01 (s : bit; xmap : std_ulogic := '0')
    return std_{0} is
  begin
    return bit_to_std(s);
  end to_01;\n""".format("ulogic"))

def disp_cond():
    w("""
  function "??" (l : std_ulogic) return boolean is
  begin
    return l = '1' or l = 'H';
  end "??";\n""")

def disp_all_norm_funcs(version):
    "Generate all function bodies for conversion"
    if version >= V08:
        disp_conv_01()
    for typ in [ "x01", "x01z", "ux01" ]:
        for v in vec_types:
            disp_conv_vec_vec(typ, v)
        disp_conv_std (typ)
        for v in vec_types:
            disp_conv_bv_vec(typ, v)
        disp_conv_b_t(typ)
    if version >= V08:
        disp_cond()

def disp_all_isx_funcs(version):
    "Generate all function bodies for isx functions"
    for v in vec_types:
        w("""
  function is_X (s : std_{0}_vector) return boolean is
  begin
    for i in s'range loop
      if std_x (s (i)) then
        return true;
      end if;
    end loop;
    return false;
  end is_X;\n""".format(v))

    w("""
  function is_X (s : std_ulogic) return boolean is
  begin
    return std_x (s);
  end is_X;\n""")


# Patterns to replace
pats = {'  @TAB\n' : disp_tables,
        '  @LOG\n' : disp_all_log_funcs,
        '  @CONV\n': disp_all_conv_funcs,
        '  @NORM\n': disp_all_norm_funcs,
        '  @ISX\n' : disp_all_isx_funcs }

spec_file='std_logic_1164.proto'
proto_file='std_logic_1164-body.proto'

def gen_body(filename, version):
    global out
    out = open(filename, 'w')
    w('--  This -*- vhdl -*- file was generated from ' + proto_file + '\n')
    keep = True
    for line in open(proto_file):
        if line in pats:
            pats[line](version)
            continue
        if line == '  @BEG V08\n':
            keep = (version == V08)
            continue
        if line == '  @END V08\n':
            keep = True
            continue
        if keep:
            w(line)
    out.close()

VDICT={'--V87':  lambda x: x == V87,
       '--!V87': lambda x: x != V87,
       '--V93':  lambda x: x == V93,
       '--V08':  lambda x: x == V08,
       '--!V08': lambda x: x != V08}

def preprocess_line(line, version):
    for p in VDICT:
        pos = line.find(p)
        if pos >= 0:
            if not VDICT[p](version):
                return None
            l = line[:pos].rstrip() + '\n'
            return l
    return line

def copy_spec(dest, version):
    out=open(dest, 'w')
    for line in open(spec_file):
        l = preprocess_line(line, version)
        if l is not None:
            out.write(l)
    out.close()

# Copy spec
copy_spec('v87/std_logic_1164.vhdl', V87)
copy_spec('v93/std_logic_1164.vhdl', V93)
copy_spec('v08/std_logic_1164.vhdl', V08)

# Generate bodies
gen_body('v87/std_logic_1164-body.vhdl', V87)

binary_funcs.append("xnor")
gen_body('v93/std_logic_1164-body.vhdl', V93)

vec_types = ['ulogic']
gen_body('v08/std_logic_1164-body.vhdl', V08)



#!/usr/bin/env python

import distutils.command.build_py
from distutils.core import setup
import re

def get_version():
    # Try from version.py.  Reads it to avoid to load the shared library.
    r = re.compile("^__version__ = '(.*)'\n")
    try:
        l = open('libghdl/version.py').read()
        m = r.match(l)
        if m:
            return m.group(1)
    except:
        pass
    raise Exception("Cannot find version")

# Extract the version now, as setup() may change the current directory.
version=get_version()

setup(
    name='libghdl',
    version=version,
    description='Interface to ghdl, a VHDL analyzer',
    long_description="""GHDL is a vhdl simulator and libghdl provides a low-level
interface to the parser. This library gives access to the AST so that you can
write tools like linters.
""",
    author='Tristan Gingold',
    author_email='tgingold@free.fr',
    url='http://github.com/ghdl/ghdl',
    license='GPL-2.0-or-later',
    package_dir={
        'libghdl': './libghdl'
    },
    packages=[
        'libghdl',
        'libghdl.thin',
        'libghdl.thin.vhdl'
    ]
)



from libghdl import libghdl
from ctypes import c_void_p

EOT = b'\x04'

No_Source_File_Entry = 0

No_Location = 0

Location_To_File = libghdl.files_map__location_to_file

Location_File_To_Pos = libghdl.files_map__location_file_to_pos

Location_File_To_Line = libghdl.files_map__location_file_to_line

Location_File_Line_To_Offset = \
    libghdl.files_map__location_file_line_to_offset

Location_File_Line_To_Col = libghdl.files_map__location_file_line_to_col

File_To_Location = libghdl.files_map__file_to_location

File_Pos_To_Location = libghdl.files_map__file_pos_to_location

File_Line_To_Position = libghdl.files_map__file_line_to_position

Get_File_Name = libghdl.files_map__get_file_name

Get_Directory_Name = libghdl.files_map__get_directory_name

Get_File_Buffer = libghdl.files_map__get_file_buffer
Get_File_Buffer.restype = c_void_p

Get_File_Length = libghdl.files_map__get_file_length
Set_File_Length = libghdl.files_map__set_file_length

Read_Source_File = libghdl.files_map__read_source_file

Reserve_Source_File = libghdl.files_map__reserve_source_file

Discard_Source_File = libghdl.files_map__discard_source_file
Free_Source_File = libghdl.files_map__free_source_file

Get_Last_Source_File_Entry = libghdl.files_map__get_last_source_file_entry



import ctypes
import os
import sys
from os.path import dirname, join, exists, normpath
from shutil import which
from libghdl.version import __version__

def _to_char_p(arg):
    return ctypes.c_char_p(arg), len(arg)


def _get_libghdl_name():
    """Get the name of the libghdl library (with version and extension)"""
    ver = __version__.replace('-', '_').replace('.', '_')
    ext = {'win32': 'dll',
           'cygwin': 'dll',
           'darwin': 'dylib'}.get(sys.platform, 'so')
    return 'libghdl-' + ver + '.' + ext


def _check_libghdl_libdir(libdir, basename):
    """Return libghdl path in :param libdir" if found or None"""
    if libdir is None:
        return None
    # print('libghdl: check in {}'.format(libdir))
    res = join(libdir, basename)
    if exists(res):
        return res
    return None


def _check_libghdl_bindir(bindir, basename):
    if bindir is None:
        return None
    return _check_libghdl_libdir(normpath(join(bindir, '..', 'lib')), basename)


def _get_libghdl_path():
    """Locate the directory where the shared library is"""
    basename = _get_libghdl_name()
    # Try GHDL_PREFIX
    r = os.environ.get('GHDL_PREFIX')
    if r is not None:
        # GHDL_PREFIX is the prefix of the vhdl libraries, so remove the
        # last path component.
        r = _check_libghdl_libdir(dirname(r), basename)
        if r is not None:
            return r
    # Try VUNIT_GHDL_PATH (path of the ghdl binary when using VUnit).
    r = _check_libghdl_bindir (os.environ.get('VUNIT_GHDL_PATH'), basename)
    if r is not None:
        return r
    # Try GHDL (name/path of the ghdl binary)
    r = os.environ.get('GHDL', 'ghdl')
    r = which(r)
    if r is not None:
        r = _check_libghdl_bindir(dirname(r), basename)
        if r is not None:
            return r
    # Try within libghdl/ python installation
    r = __file__
    r = _check_libghdl_bindir(dirname(r), basename)
    if r is not None:
        return r
    # Try when running from the build directory
    r = normpath(join(dirname(__file__), '..', '..', 'lib'))
    r = _check_libghdl_libdir(r, basename)
    if r is not None:
        return r
    # Failed.
    raise Exception('Cannot find libghdl {}'.format(basename))


# Load the shared library
_libghdl_path = _get_libghdl_path()
libghdl = ctypes.CDLL(_libghdl_path)

# Initialize it.
libghdl.libghdl_init()
libghdl.libghdl__set_hooks_for_analysis()

# Set the prefix in order to locate the vhdl libraries.
libghdl.libghdl__set_exec_prefix(
    *_to_char_p(dirname(dirname(_libghdl_path)).encode('utf-8')))

def set_option(opt):
    "Set option OPT.  Return true iff the option is known and handled"
    return libghdl.libghdl__set_option(*_to_char_p(opt)) == 0


def analyze_init():
    return libghdl.libghdl__analyze_init()


def analyze_file(fname):
    return libghdl.libghdl__analyze_file(*_to_char_p(fname))


def disp_config():
    return libghdl.ghdllocal__disp_config_prefixes()
